#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Run with:
# PYTHONPATH=.. python3 test_choice_state.py
#
"""
This test tests the ASL choice state implementation
"""

import sys
assert sys.version_info >= (3, 0) # Bomb out if not running Python3

import unittest
import json
from threading import Timer
from asl_workflow_engine.logger import init_logging
from asl_workflow_engine.state_engine import StateEngine

ASL = """{
    "Comment": "Test Choice State",
    "StartAt": "ChoiceState",
    "States": {
        "ChoiceState": {
            "Type": "Choice",
            "Choices": [
                {
                    "Variable": "$.string_equals",
                    "StringEquals": "hello world",
                    "Next": "StringEqualsSuccess"
                },
                {
                    "Variable": "$.string_equals_path",
                    "StringEqualsPath": "$.test_value",
                    "Next": "StringEqualsSuccess"
                },
                {
                    "Variable": "$.case_insensitive_string_equals",
                    "CaseInsensitiveStringEquals": "hello world",
                    "Next": "StringEqualsSuccess"
                },
                {
                    "Variable": "$.case_insensitive_string_equals_path",
                    "CaseInsensitiveStringEqualsPath": "$.test_value",
                    "Next": "StringEqualsSuccess"
                },
                {
                    "Variable": "$.string_less_than",
                    "StringLessThan": "apple",
                    "Next": "StringLessThanSuccess"
                },
                {
                    "Variable": "$.string_less_than_path",
                    "StringLessThanPath": "$.test_value",
                    "Next": "StringLessThanSuccess"
                },
                {
                    "Variable": "$.string_greater_than",
                    "StringGreaterThan": "airbnb",
                    "Next": "StringGreaterThanSuccess"
                },
                {
                    "Variable": "$.string_greater_than_path",
                    "StringGreaterThanPath": "$.test_value",
                    "Next": "StringGreaterThanSuccess"
                },
                {
                    "Variable": "$.string_less_than_equals",
                    "StringLessThanEquals": "apple",
                    "Next": "StringLessThanEqualsSuccess"
                },
                {
                    "Variable": "$.string_less_than_equals_path",
                    "StringLessThanEqualsPath": "$.test_value",
                    "Next": "StringLessThanEqualsSuccess"
                },
                {
                    "Variable": "$.string_greater_than_equals",
                    "StringGreaterThanEquals": "airbnb",
                    "Next": "StringGreaterThanEqualsSuccess"
                },
                {
                    "Variable": "$.string_greater_than_equals_path",
                    "StringGreaterThanEqualsPath": "$.test_value",
                    "Next": "StringGreaterThanEqualsSuccess"
                },
                {
                    "Variable": "$.string_matches1",
                    "StringMatches": "foo*.log",
                    "Next": "StringMatchesSuccess"
                },
                {
                    "Variable": "$.string_matches2",
                    "StringMatches": "*.log",
                    "Next": "StringMatchesSuccess"
                },
                {
                    "Variable": "$.string_matches3",
                    "StringMatches": "foo*.*",
                    "Next": "StringMatchesSuccess"
                },
                {
                    "Variable": "$.string_matches4",
                    "StringMatches": "foo\\\\*[hello]\\\\test\\\\.log",
                    "Next": "StringMatchesSuccess"
                },
                {
                    "Variable": "$.numeric_equals",
                    "NumericEquals": 1234,
                    "Next": "NumericEqualsSuccess"
                },
                {
                    "Variable": "$.numeric_equals_path",
                    "NumericEqualsPath": "$.test_value",
                    "Next": "NumericEqualsSuccess"
                },
                {
                    "Variable": "$.numeric_less_than",
                    "NumericLessThan": 1234,
                    "Next": "NumericLessThanSuccess"
                },
                {
                    "Variable": "$.numeric_less_than_path",
                    "NumericLessThanPath": "$.test_value",
                    "Next": "NumericLessThanSuccess"
                },
                {
                    "Variable": "$.numeric_greater_than",
                    "NumericGreaterThan": 1234,
                    "Next": "NumericGreaterThanSuccess"
                },
                {
                    "Variable": "$.numeric_greater_than_path",
                    "NumericGreaterThanPath": "$.test_value",
                    "Next": "NumericGreaterThanSuccess"
                },
                {
                    "Variable": "$.numeric_less_than_equals",
                    "NumericLessThanEquals": 1234,
                    "Next": "NumericLessThanEqualsSuccess"
                },
                {
                    "Variable": "$.numeric_less_than_equals_path",
                    "NumericLessThanEqualsPath": "$.test_value",
                    "Next": "NumericLessThanEqualsSuccess"
                },
                {
                    "Variable": "$.numeric_greater_than_equals",
                    "NumericGreaterThanEquals": 1234,
                    "Next": "NumericGreaterThanEqualsSuccess"
                },
                {
                    "Variable": "$.numeric_greater_than_equals_path",
                    "NumericGreaterThanEqualsPath": "$.test_value",
                    "Next": "NumericGreaterThanEqualsSuccess"
                },
                {
                    "Variable": "$.boolean_equals",
                    "BooleanEquals": true,
                    "Next": "BooleanEqualsSuccess"
                },
                {
                    "Variable": "$.boolean_equals_path",
                    "BooleanEqualsPath": "$.test_value",
                    "Next": "BooleanEqualsSuccess"
                },
                {
                    "And": [
                        {
                            "Variable": "$.and_test_value",
                            "NumericGreaterThanEquals": 20
                        },
                        {
                            "Variable": "$.and_test_value",
                            "NumericLessThan": 30
                        }
                    ],
                    "Next": "ValueInTwenties"
                },
                {
                    "Or": [
                        {
                            "Variable": "$.animal",
                            "StringEquals": "cat"
                        },
                        {
                            "Variable": "$.animal",
                            "StringEquals": "dog"
                        }
                    ],
                    "Next": "IsPet"
                },
                {
                    "And": [
                        {
                            "Or": [
                                {
                                    "Variable": "$.name",
                                    "StringEquals": "shrek"
                                },
                                {
                                    "Variable": "$.name",
                                    "StringEquals": "donkey"
                                }
                            ]
                        },
                        {
                            "And": [
                                {
                                    "Variable": "$.age",
                                    "NumericGreaterThanEquals": 40
                                },
                                {
                                    "Variable": "$.age",
                                    "NumericLessThan": 60
                                }
                            ]

                        },
                        {
                            "Not": {
                                "Variable": "$.colour",
                                "StringEquals": "pink"
                            }
                        }
                    ],
                    "Next": "IsCharacter"
                },
                {
                    "Variable": "$.timestamp_equals",
                    "TimestampEquals": "2019-08-08T10:55:25.325038+01:00",
                    "Next": "TimestampEqualsSuccess"
                },
                {
                    "Variable": "$.timestamp_equals_path",
                    "TimestampEqualsPath": "$.test_value",
                    "Next": "TimestampEqualsSuccess"
                },
                {
                    "Variable": "$.timestamp_less_than",
                    "TimestampLessThan": "2019-08-08T10:55:25.325038+01:00",
                    "Next": "TimestampLessThanSuccess"
                },
                {
                    "Variable": "$.timestamp_less_than_path",
                    "TimestampLessThanPath": "$.test_value",
                    "Next": "TimestampLessThanSuccess"
                },
                {
                    "Variable": "$.timestamp_greater_than",
                    "TimestampGreaterThan": "2019-08-08T10:55:25.325038+01:00",
                    "Next": "TimestampGreaterThanSuccess"
                },
                {
                    "Variable": "$.timestamp_greater_than_path",
                    "TimestampGreaterThanPath": "$.test_value",
                    "Next": "TimestampGreaterThanSuccess"
                },
                {
                    "Variable": "$.timestamp_less_than_equals",
                    "TimestampLessThanEquals": "2019-08-08T10:55:25.325038+01:00",
                    "Next": "TimestampLessThanEqualsSuccess"
                },
                {
                    "Variable": "$.timestamp_less_than_equals_path",
                    "TimestampLessThanEqualsPath": "$.test_value",
                    "Next": "TimestampLessThanEqualsSuccess"
                },
                {
                    "Variable": "$.timestamp_greater_than_equals",
                    "TimestampGreaterThanEquals": "2019-08-08T10:55:25.325038+01:00",
                    "Next": "TimestampGreaterThanEqualsSuccess"
                },
                {
                    "Variable": "$.timestamp_greater_than_equals_path",
                    "TimestampGreaterThanEqualsPath": "$.test_value",
                    "Next": "TimestampGreaterThanEqualsSuccess"
                },
                {
                    "Variable": "$.is_boolean",
                    "IsBoolean": true,
                    "Next": "TypeTestSuccess"
                },
                {
                    "Variable": "$.is_null",
                    "IsNull": true,
                    "Next": "TypeTestSuccess"
                },
                {
                    "Variable": "$.is_numeric",
                    "IsNumeric": true,
                    "Next": "TypeTestSuccess"
                },
                {
                    "Variable": "$.is_string",
                    "IsString": true,
                    "Next": "TypeTestSuccess"
                },
                {
                    "Variable": "$.is_present",
                    "IsPresent": true,
                    "Next": "TypeTestSuccess"
                },
                {
                    "Variable": "$.is_timestamp",
                    "IsTimestamp": true,
                    "Next": "TypeTestSuccess"
                },
                {
                    "Not": {
                        "Variable": "$.not_string_equals",
                        "StringEquals": "hello world"
                    },
                    "Next": "NotStringEqualsSuccess"
                }
            ],
            "Default": "FailState"
        },
        "FailState": {
            "Type": "Fail",
            "Error": "NoMatchError",
            "Cause": "No Matches!"
        },
        "StringEqualsSuccess": {
            "Type": "Pass",
            "End": true
        },
        "StringLessThanSuccess": {
            "Type": "Pass",
            "End": true
        },
        "StringGreaterThanSuccess": {
            "Type": "Pass",
            "End": true
        },
        "StringLessThanEqualsSuccess": {
            "Type": "Pass",
            "End": true
        },
        "StringGreaterThanEqualsSuccess": {
            "Type": "Pass",
            "End": true
        },
        "StringMatchesSuccess": {
            "Type": "Pass",
            "End": true
        },
        "NumericEqualsSuccess": {
            "Type": "Pass",
            "End": true
        },
        "NumericLessThanSuccess": {
            "Type": "Pass",
            "End": true
        },
        "NumericGreaterThanSuccess": {
            "Type": "Pass",
            "End": true
        },
        "NumericLessThanEqualsSuccess": {
            "Type": "Pass",
            "End": true
        },
        "NumericGreaterThanEqualsSuccess": {
            "Type": "Pass",
            "End": true
        },
        "BooleanEqualsSuccess": {
            "Type": "Pass",
            "End": true
        },
        "ValueInTwenties": {
            "Type": "Pass",
            "End": true
        },
        "IsPet": {
            "Type": "Pass",
            "End": true
        },
        "IsCharacter": {
            "Type": "Pass",
            "End": true
        },
        "TimestampEqualsSuccess": {
            "Type": "Pass",
            "End": true
        },
        "TimestampLessThanSuccess": {
            "Type": "Pass",
            "End": true
        },
        "TimestampGreaterThanSuccess": {
            "Type": "Pass",
            "End": true
        },
        "TimestampLessThanEqualsSuccess": {
            "Type": "Pass",
            "End": true
        },
        "TimestampGreaterThanEqualsSuccess": {
            "Type": "Pass",
            "End": true
        },
        "NotStringEqualsSuccess": {
            "Type": "Pass",
            "End": true
        },
        "TypeTestSuccess": {
            "Type": "Pass",
            "End": true
        }
    }
}"""


context = '{"StateMachine": {"Id": "arn:aws:states:local:0123456789:stateMachine:simple_state_machine", "Definition": ' + ASL + '}}'

"""
Create a simple EventDispatcher stub so that we can test the State Engine
without requiring the messaging fabric. Rather than publishing each new state
to a queue the stub simply calls notify on the StateEngine with the new state
information. This simplistic approach should be OK for most tests, but we should
be careful of issues due to recursion so we may need to revisit this IDC.
"""
class EventDispatcherStub(object):

    def __init__(self, state_engine, config):

        """
        Create an association with the state engine and give that a reference
        back to this event dispatcher so that it can publish events and make
        use of the set_timeout time scheduler.
        """
        self.state_engine = state_engine
        self.state_engine.event_dispatcher = self
        self.message_count = -1
        self.last_event = None # Retain last event so test can check its value

    """
    This simple threaded timeout should work OK, the real timeout is actually
    implemented using Pika's connection.call_later() which is single threaded
    and handled within Pika's event loop. That approach plays much better with
    Pika's event loop.
    """
    def set_timeout(self, callback, delay):
        t = Timer(delay/1000, callback)
        t.start()
        return t

    def dispatch(self, message):
        """
        Start at -1 and increment before call to notify as this stub, unlike
        the real EventDispatcher will recursively call dispatch as the State
        Engine calls publish, this means code after the call to notify won't
        be reached when one might expect it to.
        """
        self.message_count += 1
        # The notify method expects a JSON object not a string.
        self.state_engine.notify(json.loads(message), self.message_count)

    def acknowledge(self, id):
        pass

    def publish(self, item):
        # Convert event back to JSON string for dispatching.
        self.last_event = item
        self.dispatch(json.dumps(item))
    
    def broadcast(self, subject, message, carrier_properties=None):
        pass

"""
This stubs out the real TaskDispatcher execute_task method which requires
messaging infrastructure to run whereas this test is just a state machine test.
"""
def execute_task_stub(resource_arn, parameters, callback):
    name = resource_arn.split(":")[-1]
    result = {"reply": name + " reply"}
    callback(result)


class TestChoiceState(unittest.TestCase):

    def setUp(self):
        # Initialise logger
        logger = init_logging(log_name="test_choice_state")
        config = {
            "state_engine": {
                "store_url": "ASL_store.json", 
                "execution_ttl": 500
            }
        }

        state_engine = StateEngine(config)
        # Stub out the real TaskDispatcher execute_task
        state_engine.task_dispatcher.execute_task = execute_task_stub
        self.event_dispatcher = EventDispatcherStub(state_engine, config)

    #---------- String ---------------------------------------------------------
    def test_string_equals(self):
        print("---------- test_string_equals ----------")
        item = '{"data": {"string_equals":"hello world"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["string_equals"], "hello world")
        self.assertEqual(state, "StringEqualsSuccess")

    def test_string_equals_path(self):
        print("---------- test_string_equals_path ----------")
        item = '{"data": {"string_equals_path":"hello world", "test_value":"hello world"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["string_equals_path"], "hello world")
        self.assertEqual(state, "StringEqualsSuccess")

    def test_case_insensitive_string_equals(self):
        # Case insensitive string comparison is not covered in the ASL spec.
        # but as it is useful and trivial to implement we may as well add it.
        print("---------- test_case_insensitive_string_equals ----------")
        item = '{"data": {"case_insensitive_string_equals":"Hello world"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["case_insensitive_string_equals"], "Hello world")
        self.assertEqual(state, "StringEqualsSuccess")

    def test_case_insensitive_string_equals_path(self):
        # Case insensitive string comparison is not covered in the ASL spec.
        # but as it is useful and trivial to implement we may as well add it.
        print("---------- test_case_insensitive_string_equals_path ----------")
        item = '{"data": {"case_insensitive_string_equals_path":"Hello world", "test_value":"hello world"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["case_insensitive_string_equals_path"], "Hello world")
        self.assertEqual(state, "StringEqualsSuccess")

    def test_string_less_than(self):
        print("---------- test_string_less_than ----------")
        item = '{"data": {"string_less_than":"amazon"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["string_less_than"], "amazon")
        self.assertEqual(state, "StringLessThanSuccess")

    def test_string_less_than_path(self):
        print("---------- test_string_less_than_path ----------")
        item = '{"data": {"string_less_than_path":"amazon", "test_value":"apple"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["string_less_than_path"], "amazon")
        self.assertEqual(state, "StringLessThanSuccess")

    def test_string_greater_than(self):
        print("---------- test_string_greater_than ----------")
        item = '{"data": {"string_greater_than":"amazon"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["string_greater_than"], "amazon")
        self.assertEqual(state, "StringGreaterThanSuccess")

    def test_string_greater_than_path(self):
        print("---------- test_string_greater_than_path ----------")
        item = '{"data": {"string_greater_than_path":"amazon", "test_value":"airbnb"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["string_greater_than_path"], "amazon")
        self.assertEqual(state, "StringGreaterThanSuccess")

    def test_string_less_than_equals(self):
        print("---------- test_string_less_than_equals ----------")
        item = '{"data": {"string_less_than_equals":"amazon"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["string_less_than_equals"], "amazon")
        self.assertEqual(state, "StringLessThanEqualsSuccess")

        item = '{"data": {"string_less_than_equals":"apple"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["string_less_than_equals"], "apple")
        self.assertEqual(state, "StringLessThanEqualsSuccess")

    def test_string_less_than_equals_path(self):
        print("---------- test_string_less_than_equals_path ----------")
        item = '{"data": {"string_less_than_equals_path":"amazon", "test_value":"apple"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["string_less_than_equals_path"], "amazon")
        self.assertEqual(state, "StringLessThanEqualsSuccess")

        item = '{"data": {"string_less_than_equals_path":"apple", "test_value":"apple"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["string_less_than_equals_path"], "apple")
        self.assertEqual(state, "StringLessThanEqualsSuccess")
    
    def test_string_greater_than_equals(self):
        print("---------- test_string_greater_than_equals ----------")
        item = '{"data": {"string_greater_than_equals":"amazon"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["string_greater_than_equals"], "amazon")
        self.assertEqual(state, "StringGreaterThanEqualsSuccess")

        item = '{"data": {"string_greater_than_equals":"airbnb"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["string_greater_than_equals"], "airbnb")
        self.assertEqual(state, "StringGreaterThanEqualsSuccess")

    def test_string_greater_than_equals_path(self):
        print("---------- test_string_greater_than_equals_path ----------")
        item = '{"data": {"string_greater_than_equals_path":"amazon", "test_value":"airbnb"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["string_greater_than_equals_path"], "amazon")
        self.assertEqual(state, "StringGreaterThanEqualsSuccess")

        item = '{"data": {"string_greater_than_equals_path":"airbnb", "test_value":"airbnb"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["string_greater_than_equals_path"], "airbnb")
        self.assertEqual(state, "StringGreaterThanEqualsSuccess")

    def test_string_matches(self):
        print("---------- test_string_matches ----------")
        item = '{"data": {"string_matches1":"foo23.log"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["string_matches1"], "foo23.log")
        self.assertEqual(state, "StringMatchesSuccess")

        item = '{"data": {"string_matches1":"fool.log"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["string_matches1"], "fool.log")
        self.assertEqual(state, "StringMatchesSuccess")

        item = '{"data": {"string_matches2":"foo23.log"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["string_matches2"], "foo23.log")
        self.assertEqual(state, "StringMatchesSuccess")

        item = '{"data": {"string_matches2":"zebra.log"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["string_matches2"], "zebra.log")
        self.assertEqual(state, "StringMatchesSuccess")

        item = '{"data": {"string_matches3":"foo23.log"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["string_matches3"], "foo23.log")
        self.assertEqual(state, "StringMatchesSuccess")

        item = '{"data": {"string_matches3":"foobar.zebra"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["string_matches3"], "foobar.zebra")
        self.assertEqual(state, "StringMatchesSuccess")

        # Test escape characters in value to match literal chars present in variable
        # In JSON the escaped string \\\\ represents \
        item = '{"data": {"string_matches4":"foo*[hello]\\\\test\\\\.log"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["string_matches4"], "foo*[hello]\\test\\.log")
        self.assertEqual(state, "StringMatchesSuccess")

    #---------- Numeric --------------------------------------------------------

    def test_numeric_equals(self):
        print("---------- test_numeric_equals ----------")
        item = '{"data": {"numeric_equals":1234}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["numeric_equals"], 1234)
        self.assertEqual(state, "NumericEqualsSuccess")

    def test_numeric_equals_path(self):
        print("---------- test_numeric_equals_path ----------")
        item = '{"data": {"numeric_equals_path":1234, "test_value":1234}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["numeric_equals_path"], 1234)
        self.assertEqual(state, "NumericEqualsSuccess")
    
    def test_numeric_less_than(self):
        print("---------- test_numeric_less_than ----------")
        item = '{"data": {"numeric_less_than":123}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["numeric_less_than"], 123)
        self.assertEqual(state, "NumericLessThanSuccess")

    def test_numeric_less_than_path(self):
        print("---------- test_numeric_less_than_path ----------")
        item = '{"data": {"numeric_less_than_path":123, "test_value":1234}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["numeric_less_than_path"], 123)
        self.assertEqual(state, "NumericLessThanSuccess")
    
    def test_numeric_greater_than(self):
        print("---------- test_numeric_greater_than ----------")
        item = '{"data": {"numeric_greater_than":12345}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["numeric_greater_than"], 12345)
        self.assertEqual(state, "NumericGreaterThanSuccess")

    def test_numeric_greater_than_path(self):
        print("---------- test_numeric_greater_than_path ----------")
        item = '{"data": {"numeric_greater_than_path":12345, "test_value":1234}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["numeric_greater_than_path"], 12345)
        self.assertEqual(state, "NumericGreaterThanSuccess")
    
    def test_numeric_less_than_equals(self):
        print("---------- test_numeric_less_than_equals ----------")
        item = '{"data": {"numeric_less_than_equals":1234}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["numeric_less_than_equals"], 1234)
        self.assertEqual(state, "NumericLessThanEqualsSuccess")

        item = '{"data": {"numeric_less_than_equals":123}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["numeric_less_than_equals"], 123)
        self.assertEqual(state, "NumericLessThanEqualsSuccess")

    def test_numeric_less_than_equals_path(self):
        print("---------- test_numeric_less_than_equals_path ----------")
        item = '{"data": {"numeric_less_than_equals_path":1234, "test_value":1234}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["numeric_less_than_equals_path"], 1234)
        self.assertEqual(state, "NumericLessThanEqualsSuccess")

        item = '{"data": {"numeric_less_than_equals_path":123, "test_value":1234}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["numeric_less_than_equals_path"], 123)
        self.assertEqual(state, "NumericLessThanEqualsSuccess")
    
    def test_numeric_greater_than_equals(self):
        print("---------- test_numeric_greater_than_equals ----------")
        item = '{"data": {"numeric_greater_than_equals":1234}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["numeric_greater_than_equals"], 1234)
        self.assertEqual(state, "NumericGreaterThanEqualsSuccess")

        item = '{"data": {"numeric_greater_than_equals":12345}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["numeric_greater_than_equals"], 12345)
        self.assertEqual(state, "NumericGreaterThanEqualsSuccess")

    def test_numeric_greater_than_equals_path(self):
        print("---------- test_numeric_greater_than_equals_path ----------")
        item = '{"data": {"numeric_greater_than_equals_path":1234, "test_value":1234}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["numeric_greater_than_equals_path"], 1234)
        self.assertEqual(state, "NumericGreaterThanEqualsSuccess")

        item = '{"data": {"numeric_greater_than_equals_path":12345, "test_value":1234}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["numeric_greater_than_equals_path"], 12345)
        self.assertEqual(state, "NumericGreaterThanEqualsSuccess")

    #---------- Boolean --------------------------------------------------------
    
    def test_boolean_equals(self):
        print("---------- test_boolean_equals ----------")
        # Note difference between JSON true and Python True
        item = '{"data": {"boolean_equals":true}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["boolean_equals"], True)
        self.assertEqual(state, "BooleanEqualsSuccess")

    def test_boolean_equals_path(self):
        print("---------- test_boolean_equals_path ----------")
        # Note difference between JSON true and Python True
        item = '{"data": {"boolean_equals_path":true, "test_value":true}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["boolean_equals_path"], True)
        self.assertEqual(state, "BooleanEqualsSuccess")

    def test_and(self):
        print("---------- test_and ----------")
        item = '{"data": {"and_test_value":22}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["and_test_value"], 22)
        self.assertEqual(state, "ValueInTwenties")

    def test_or(self):
        print("---------- test_or ----------")
        item = '{"data": {"animal":"cat"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["animal"], "cat")
        self.assertEqual(state, "IsPet")

        item = '{"data": {"animal":"dog"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["animal"], "dog")
        self.assertEqual(state, "IsPet")

    def test_complex_logic(self):
        print("---------- test_complex_logic ----------")
        item = '{"data": {"name":"shrek", "age":55, "colour":"green"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["name"], "shrek")
        self.assertEqual(state, "IsCharacter")

        item = '{"data": {"name":"donkey", "age":50, "colour":"grey"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["name"], "donkey")
        self.assertEqual(state, "IsCharacter")

    def test_not_string_equals(self):
        # In other words this is the != operator
        print("---------- test_not_string_equals ----------")
        item = '{"data": {"not_string_equals":"not hello world"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["not_string_equals"], "not hello world")
        self.assertEqual(state, "NotStringEqualsSuccess")

    def test_timestamp_equals(self):
        print("---------- test_timestamp_equals ----------")
        item = '{"data": {"timestamp_equals":"2019-08-08T10:55:25.325038+01:00"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["timestamp_equals"], "2019-08-08T10:55:25.325038+01:00")
        self.assertEqual(state, "TimestampEqualsSuccess")

        # This is the same time but represented in Zulu so test that matches too.
        item = '{"data": {"timestamp_equals":"2019-08-08T09:55:25.325038Z"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["timestamp_equals"], "2019-08-08T09:55:25.325038Z")
        self.assertEqual(state, "TimestampEqualsSuccess")

    def test_timestamp_equals_path(self):
        print("---------- test_timestamp_equals_path ----------")
        item = '{"data": {"timestamp_equals_path":"2019-08-08T10:55:25.325038+01:00", "test_value":"2019-08-08T10:55:25.325038+01:00"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["timestamp_equals_path"], "2019-08-08T10:55:25.325038+01:00")
        self.assertEqual(state, "TimestampEqualsSuccess")

        # This is the same time but represented in Zulu so test that matches too.
        item = '{"data": {"timestamp_equals_path":"2019-08-08T09:55:25.325038Z", "test_value":"2019-08-08T10:55:25.325038+01:00"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["timestamp_equals_path"], "2019-08-08T09:55:25.325038Z")
        self.assertEqual(state, "TimestampEqualsSuccess")
    
    def test_timestamp_less_than(self):
        print("---------- test_timestamp_less_than ----------")
        item = '{"data": {"timestamp_less_than":"2019-08-08T09:55:25.325038+01:00"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["timestamp_less_than"], "2019-08-08T09:55:25.325038+01:00")
        self.assertEqual(state, "TimestampLessThanSuccess")
    
    def test_timestamp_less_than_path(self):
        print("---------- test_timestamp_less_than_path ----------")
        item = '{"data": {"timestamp_less_than_path":"2019-08-08T09:55:25.325038+01:00", "test_value":"2019-08-08T10:55:25.325038+01:00"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["timestamp_less_than_path"], "2019-08-08T09:55:25.325038+01:00")
        self.assertEqual(state, "TimestampLessThanSuccess")

    def test_timestamp_greater_than(self):
        print("---------- test_timestamp_greater_than ----------")
        item = '{"data": {"timestamp_greater_than":"2019-08-08T11:55:25.325038+01:00"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["timestamp_greater_than"], "2019-08-08T11:55:25.325038+01:00")
        self.assertEqual(state, "TimestampGreaterThanSuccess")

    def test_timestamp_greater_than_path(self):
        print("---------- test_timestamp_greater_than_path ----------")
        item = '{"data": {"timestamp_greater_than_path":"2019-08-08T11:55:25.325038+01:00", "test_value":"2019-08-08T10:55:25.325038+01:00"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["timestamp_greater_than_path"], "2019-08-08T11:55:25.325038+01:00")
        self.assertEqual(state, "TimestampGreaterThanSuccess")
    
    def test_timestamp_less_than_equals(self):
        print("---------- test_timestamp_less_than_equals ----------")
        item = '{"data": {"timestamp_less_than_equals":"2019-08-08T10:55:25.325038+01:00"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["timestamp_less_than_equals"], "2019-08-08T10:55:25.325038+01:00")
        self.assertEqual(state, "TimestampLessThanEqualsSuccess")

        item = '{"data": {"timestamp_less_than_equals":"2019-08-08T09:55:25.325038+01:00"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["timestamp_less_than_equals"], "2019-08-08T09:55:25.325038+01:00")
        self.assertEqual(state, "TimestampLessThanEqualsSuccess")

    def test_timestamp_less_than_equals_path(self):
        print("---------- test_timestamp_less_than_equals_path ----------")
        item = '{"data": {"timestamp_less_than_equals_path":"2019-08-08T10:55:25.325038+01:00", "test_value":"2019-08-08T10:55:25.325038+01:00"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["timestamp_less_than_equals_path"], "2019-08-08T10:55:25.325038+01:00")
        self.assertEqual(state, "TimestampLessThanEqualsSuccess")

        item = '{"data": {"timestamp_less_than_equals_path":"2019-08-08T09:55:25.325038+01:00", "test_value":"2019-08-08T10:55:25.325038+01:00"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["timestamp_less_than_equals_path"], "2019-08-08T09:55:25.325038+01:00")
        self.assertEqual(state, "TimestampLessThanEqualsSuccess")

    def test_timestamp_greater_than_equals(self):
        print("---------- test_timestamp_greater_than_equals ----------")
        item = '{"data": {"timestamp_greater_than_equals":"2019-08-08T10:55:25.325038+01:00"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["timestamp_greater_than_equals"], "2019-08-08T10:55:25.325038+01:00")
        self.assertEqual(state, "TimestampGreaterThanEqualsSuccess")

        item = '{"data": {"timestamp_greater_than_equals":"2019-08-08T11:55:25.325038+01:00"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["timestamp_greater_than_equals"], "2019-08-08T11:55:25.325038+01:00")
        self.assertEqual(state, "TimestampGreaterThanEqualsSuccess")

    def test_timestamp_greater_than_equals_path(self):
        print("---------- test_timestamp_greater_than_equals_path ----------")
        item = '{"data": {"timestamp_greater_than_equals_path":"2019-08-08T10:55:25.325038+01:00", "test_value":"2019-08-08T10:55:25.325038+01:00"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["timestamp_greater_than_equals_path"], "2019-08-08T10:55:25.325038+01:00")
        self.assertEqual(state, "TimestampGreaterThanEqualsSuccess")

        item = '{"data": {"timestamp_greater_than_equals_path":"2019-08-08T11:55:25.325038+01:00", "test_value":"2019-08-08T10:55:25.325038+01:00"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["timestamp_greater_than_equals_path"], "2019-08-08T11:55:25.325038+01:00")
        self.assertEqual(state, "TimestampGreaterThanEqualsSuccess")

    #---------- Type Test ------------------------------------------------------

    def test_is_boolean(self):
        print("---------- test_is_boolean ----------")
        item = '{"data": {"is_boolean":false}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["is_boolean"], False)
        self.assertEqual(state, "TypeTestSuccess")

    def test_is_null(self):
        print("---------- test_is_null ----------")
        item = '{"data": {"is_null":null}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["is_null"], None)
        self.assertEqual(state, "TypeTestSuccess")

    def test_is_numeric(self):
        print("---------- test_is_numeric ----------")
        item = '{"data": {"is_numeric":1234}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["is_numeric"], 1234)
        self.assertEqual(state, "TypeTestSuccess")

        item = '{"data": {"is_numeric":1234.5}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["is_numeric"], 1234.5)
        self.assertEqual(state, "TypeTestSuccess")

    def test_is_string(self):
        print("---------- test_is_string ----------")
        item = '{"data": {"is_string":"cat"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["is_string"], "cat")
        self.assertEqual(state, "TypeTestSuccess")

    def test_is_present(self):
        print("---------- test_is_present ----------")
        item = '{"data": {"is_present":{}}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["is_present"], {})
        self.assertEqual(state, "TypeTestSuccess")

        item = '{"data": {"is_present":false}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["is_present"], False)
        self.assertEqual(state, "TypeTestSuccess")

    def test_is_timestamp(self):
        print("---------- test_is_timestamp ----------")
        item = '{"data": {"is_timestamp":"2019-08-08T10:55:25.325038+01:00"}, "context": ' + context + '}'
        self.event_dispatcher.dispatch(item)
        data = self.event_dispatcher.last_event["data"]
        state = self.event_dispatcher.last_event["context"]["State"]["Name"]
        self.assertEqual(data["is_timestamp"], "2019-08-08T10:55:25.325038+01:00")
        self.assertEqual(state, "TypeTestSuccess")

if __name__ == '__main__':
    unittest.main()




#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Run with:
# PYTHONPATH=.. python3 step_by_step.py
#
"""
This test illustrates one step function invoking another.

This is a relatively recent AWS addition (September 2019) described here:
https://docs.aws.amazon.com/step-functions/latest/dg/connect-stepfunctions.html
https://docs.aws.amazon.com/step-functions/latest/dg/concepts-nested-workflows.html

Because it's relatively new be aware that a lot of the information online
relating to nesting stepfunctions illustrates the use of a lambda as a proxy
for example this AWS tutorial:
https://docs.aws.amazon.com/step-functions/latest/dg/tutorial-continue-new.html
Using a lambda to proxy child stepfunctions still works and is still a
legitimate approach, but it is no longer neccesary and adds an additional cost.


This test uses the AWS python SDK boto3 to access our local ASL Workflow Engine
https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/stepfunctions.html
https://docs.aws.amazon.com/code-samples/latest/catalog/code-catalog-python-example_code-stepfunctions.html
"""

import sys
assert sys.version_info >= (3, 0) # Bomb out if not running Python3

import boto3, time
from botocore.exceptions import ClientError

from asl_workflow_engine.logger import init_logging
from asl_workflow_engine.open_tracing_factory import create_tracer

"""
In the Resource ARN below the execution Name is not specified, so a UUID will
be automatically assigned. The way to specify specific execution names (if so
desired) is to pass the execution name in the stepfunction input and extract it
in the Parameter's JSONPath processing e.g. something like:
"Name.$": "$.executionName"
"""
caller_ASL = """{
    "StartAt": "StepFunctionLauncher",
    "States": {
        "StepFunctionLauncher": {
            "Type": "Task",
            "Resource": "arn:aws:states:local:0123456789:states:startExecution",
            "Parameters": {  
                "Input.$": "$",
                "StateMachineArn": "arn:aws:states:local:0123456789:stateMachine:simple_state_machine"
            },
            "End": true
        }
    }
}"""

callee_ASL = """{
    "Comment": "Test Step Function",
    "StartAt": "StartState",
    "States": {
        "StartState": {
            "Type": "Pass",
            "Next": "ChoiceState"
        },
        "ChoiceState": {
            "Type": "Choice",
            "Choices": [
                {
                    "Variable": "$.lambda",
                    "StringEquals": "InternalErrorNotHandled",
                    "Next": "InternalErrorNotHandledLambda"
                },
                {
                    "Variable": "$.lambda",
                    "StringEquals": "InternalErrorHandled",
                    "Next": "InternalErrorHandledLambda"
                },
                {
                    "Variable": "$.lambda",
                    "StringEquals": "Success",
                    "Next": "SuccessLambda"
                },
                {
                    "Variable": "$.lambda",
                    "StringEquals": "Timeout",
                    "Next": "TimeoutLambda"
                }
            ],
            "Default": "FailState"
        },
        "FailState": {
            "Type": "Fail",
            "Error": "NoLambdaError",
            "Cause": "No Matches!"
        },
        "SuccessLambda": {
            "Type": "Task",
            "Resource": "arn:aws:rpcmessage:local::function:SuccessLambda",
            "Next": "WaitState"
        },
        "InternalErrorNotHandledLambda": {
            "Type": "Task",
            "Resource": "arn:aws:rpcmessage:local::function:InternalErrorNotHandledLambda",
            "Next": "EndState"
        },
        "InternalErrorHandledLambda": {
            "Type": "Task",
            "Resource": "arn:aws:rpcmessage:local::function:InternalErrorHandledLambda",
            "Next": "EndState"
        },
        "TimeoutLambda": {
            "Type": "Task",
            "Resource": "arn:aws:rpcmessage:local::function:TimeoutLambda",
            "Next": "EndState"
        },
        "EndState": {
            "Type": "Pass",
            "End": true
        },
        "WaitState": {
            "Type": "Wait",
            "Seconds":10,
            "Next": "EndState"
        }
    }
}"""

items = ['{"lambda":"Success"}',
         '{"lambda":"InternalErrorNotHandled"}',
         '{"lambda":"InternalErrorHandled"}',
         '{"lambda":"Timeout"}']

items = ['{"lambda":"Success"}']
#items = ['{"lambda":"InternalErrorNotHandled"}']
#items = ['{"lambda":"InternalErrorHandled"}']
#items = ['{"lambda":"Timeout"}']

if __name__ == '__main__':
    # Initialise logger
    logger = init_logging(log_name='step_by_step')

    # Initialising OpenTracing. It's important to do this before the boto3.client
    # call as create_tracer "patches" boto3 to add the OpenTracing hooks.
    create_tracer("step_by_step", {"implementation": "Jaeger"})

    # Initialise the boto3 client setting the endpoint_url to our local
    # ASL Workflow Engine
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html#boto3.session.Session.client
    sfn = boto3.client("stepfunctions", endpoint_url="http://localhost:4584")

    caller_state_machine_arn = "arn:aws:states:local:0123456789:stateMachine:caller_state_machine"
    state_machine_arn = "arn:aws:states:local:0123456789:stateMachine:simple_state_machine"

    def create_state_machines():
        # Create state machine using a dummy roleArn. If it already exists an
        # exception will be thrown, we ignore that but raise other exceptions.
        try:
            response = sfn.create_state_machine(
                name="caller_state_machine", definition=caller_ASL,
                roleArn="arn:aws:iam::0123456789:role/service-role/MyRole"
            )

            response = sfn.create_state_machine(
                name="simple_state_machine", definition=callee_ASL,
                roleArn="arn:aws:iam::0123456789:role/service-role/MyRole"
            )
        except sfn.exceptions.StateMachineAlreadyExists as e:
            #print(e.response)

            response = sfn.update_state_machine(
                stateMachineArn=caller_state_machine_arn,
                definition=caller_ASL
            )

            response = sfn.update_state_machine(
                stateMachineArn=state_machine_arn,
                definition=callee_ASL
            )
        except ClientError as e:
            logger.error(e)


    create_state_machines()

    # Loop through items invoking a new state machine execution for each item
    for item in items:
        try:
            response = sfn.start_execution(
                stateMachineArn=caller_state_machine_arn,
                #name=EXECUTION_NAME, # If not specified a UUID is assigned
                input=item
            )

            """
            Get ARN of the execution just invoked on the caller_state_machine.
            Note that it's more awkward trying to get the execution ARN of
            the execution that this invokes on the callee because we let it use
            a system assigned ARN, but hey! illustrating this stuff is kind of
            the point of this example.

            Note too that, as a distributed and clustered service, the call to      
            sfn.get_execution_history() can fail even when called immediately
            after sfn.start_execution(). Here we simply wrap in a loop and if an  
            ExecutionDoesNotExist exception occurs we briefly sleep then retry.
            If polling in this way is undesireable then using notification
            events (or on real AWS StepFunctions using CloudWatch events) may
            be preferable, albeit at the cost of additional complexity.
            """
            execution_arn = response["executionArn"]
            while True:
                try:
                    history = sfn.get_execution_history(
                        executionArn=execution_arn
                    )
                    break
                except sfn.exceptions.ExecutionDoesNotExist as e:
                    time.sleep(0.1)  # Sleep for 100ms then retry

            print("Execution history for Launcher state machine execution:")
            print(execution_arn)
            print()
            print(history)
            print()

            """
            List all executions for the callee state machine. Do NOT
            use the "SUCCEEDED" statusFilter, because this example includes a
            Wait state which waits for around 10s, so the child execution will
            still only be in the RUNNING state by this point.
            """
            executions = sfn.list_executions(
                stateMachineArn=state_machine_arn
            )
            #print(executions)

        except sfn.exceptions.StateMachineDoesNotExist as e:
            logger.info(e)

            create_state_machines()
            response = sfn.start_execution(
                stateMachineArn=caller_state_machine_arn,
                #name=EXECUTION_NAME, # If not specified a UUID is assigned
                input=item
            )
        except ClientError as e:
            logger.error(e)

    time.sleep(1)  # Give OpenTracing a chance to flush buffer before exiting




#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

"""
-------------------------------- READ ME FIRST ---------------------------------
Note that in JSON and dict manipulation herein there may be a mix of camel case
and fields starting with capitals. This is unfortunate, but somewhat deliberate
as we are trying to follow the patterns used in real AWS Step Functions, which
seems to use camel case in the REST API calls but in the Context object and
indeed in the ASL specification the fields start with capitals. Be aware of this
if suddenly overcome by the urge to "make everything consistent"
--------------------------------------------------------------------------------
"""

"""
This implements the REST API for the ASL Workflow Engine. The intention is to
implement the AWS Step Functions API as described in the AWS documentation:
https://docs.aws.amazon.com/step-functions/latest/apireference/API_Operations.html

By implementing the AWS REST API semantics it becomes possible to use Amazon's
CLI and SDK so applications can use this ASL Workflow Engine as an alternative
to Amazon's for scenarios such as hybrid cloud workloads.


Example actions using AWS CLI:

# List state machines
aws stepfunctions --endpoint http://localhost:4584 list-state-machines --max-results 20

# Create a new state machine
aws stepfunctions --endpoint http://localhost:4584 create-state-machine --name my-state-machine --definition '{"Comment":"A Hello World example of the Amazon States Language using a Pass state","StartAt":"HelloWorld","States":{"HelloWorld":{"Type":"Pass","End":true}}}' --role-arn arn:aws:iam::0123456789:role/service-role/MyRole

# Create a new state machine from a file
aws stepfunctions --endpoint http://localhost:4584 create-state-machine --name simple_state_machine --definition file://<path-to-ASL-JSON> --role-arn arn:aws:iam::0123456789:role/service-role/MyRole

# Update a state machine
aws stepfunctions --endpoint http://localhost:4584 update-state-machine --definition '{"Comment":"A Hello World example of the Amazon States Language using a Pass state","StartAt":"HelloWorld","States":{"HelloWorld":{"Type":"Pass","End":true}}}' --role-arn arn:aws:iam::0123456789:role/service-role/MyRole --state-machine-arn arn:aws:states:local:0123456789:stateMachine:my-state-machine

# Describe state machine
aws stepfunctions --endpoint http://localhost:4584 describe-state-machine --state-machine-arn arn:aws:states:local:0123456789:stateMachine:my-state-machine

# Delete state machine
aws stepfunctions --endpoint http://localhost:4584 delete-state-machine --state-machine-arn arn:aws:states:local:0123456789:stateMachine:my-state-machine

# Start state machine execution
aws stepfunctions --endpoint http://localhost:4584 start-execution --state-machine-arn arn:aws:states:local:0123456789:stateMachine:my-state-machine --name my-execution --input '{"comment":"I am a great input !"}'

# List state machine executions
aws stepfunctions --endpoint http://localhost:4584 list-executions --state-machine-arn arn:aws:states:local:0123456789:stateMachine:my-state-machine

# Describe execution
aws stepfunctions --endpoint http://localhost:4584 describe-execution --execution-arn arn:aws:states:local:0123456789:execution:my-state-machine:my-execution

# Describe state machine related to execution
aws stepfunctions --endpoint http://localhost:4584 describe-state-machine-for-execution --execution-arn arn:aws:states:local:0123456789:execution:my-state-machine:my-execution

# Get execution history
aws stepfunctions --endpoint http://localhost:4584 get-execution-history --execution-arn arn:aws:states:local:0123456789:execution:my-state-machine:my-execution
"""

import sys
assert sys.version_info >= (3, 0)  # Bomb out if not running Python3

import re, time, uuid, logging, opentracing
from datetime import datetime, timezone
from flask import Flask, escape, request, jsonify, abort

from asl_workflow_engine.logger import init_logging
from asl_workflow_engine.open_tracing_factory import span_context, inject_span
from asl_workflow_engine.arn import *
from asl_workflow_engine.state_engine import MAX_DATA_LENGTH, MAX_STATE_MACHINE_LENGTH

try:  # Attempt to use ujson if available https://pypi.org/project/ujson/
    import ujson as json
except:  # Fall back to standard library json
    import json


def valid_name(name):
    return (
        isinstance(name, str)
        and len(name) > 0
        and len(name) < 81
        and not re.search(r"^.*[ <>{}[\]?*\"#%\\^|~`$&,;:/].*$", name)
    )

def valid_role_arn(arn):
    return (
        isinstance(arn, str)
        and len(arn) > 0
        and len(arn) < 257
        and re.search(r"^arn:aws:iam::[0-9]+:role\/.+$", arn)
    )

def valid_state_machine_arn(arn):
    return (
        isinstance(arn, str)
        and len(arn) > 0
        and len(arn) < 257
        and re.search(r"^arn:aws:states:.+:[0-9]+:stateMachine:.+$", arn)
    )

def valid_execution_arn(arn):
    return (
        isinstance(arn, str)
        and len(arn) > 0
        and len(arn) < 257
        and re.search(r"^arn:aws:states:.+:[0-9]+:execution:.+$", arn)
    )

def aws_error(code, message=None):
    """
    Boiler plate to return errors in the correct form for the SDKs to throw
    the expected exceptions. The format doesn't seem to be documented anywhere
    so this was grokked by looking at the botocore source code in
    https://github.com/boto/botocore/blob/develop/botocore/parsers.py
    BaseJSONParser._do_error_parse(self, response, shape)
    """
    return jsonify({
        "__type": code,
        "message": message if message else code,
    })

class RestAPI(object):
    def __init__(self, state_engine, event_dispatcher, config):
        """
        """
        self.logger = init_logging(log_name="asl_workflow_engine")
        self.logger.info("Creating {}.RestAPI, using {} JSON parser".format(
            __name__, json.__name__
        ))

        config = config.get("rest_api")
        if config:
            self.host = config.get("host", "0.0.0.0")
            self.port = config.get("port", 4584)
            self.region = config.get("region", "local")

        self.asl_store = state_engine.asl_store
        self.executions = state_engine.executions
        self.execution_history = state_engine.execution_history
        self.event_dispatcher = event_dispatcher

    def create_app(self):
        app = Flask(__name__)

        # Turn off Flask standard logging
        app.logger.disabled = True
        log = logging.getLogger("werkzeug")
        log.disabled = True

        """
        Flask "catch-all" URL
        see https://gist.github.com/fitiavana07/bf4eb97b20bbe3853681e153073c0e5e
        """
        @app.route("/", defaults={"path": ""}, methods=["POST"])
        @app.route("/<path:path>", methods=["POST"])
        def handle_post(path):
            """
            Perform initial validation of the HTTP request. The AWS Step 
            Functions API is a slightly "weird" REST API as it mostly seems to
            rely on POST and rather than using HTTP resources it uses the
            x-amz-target header to specify the action to be performed.
            """
            if not request.content_type == "application/x-amz-json-1.0":
                return "Unexpected Content-Type {}".format(request.content_type), 400

            target = request.headers.get("x-amz-target")
            if not target:
                return "Missing header x-amz-target", 400
            if not target.startswith("AWSStepFunctions."):
                return "Malformed header x-amz-target", 400

            action = target.split(".")[1]
            # print(action)

            data = request.data

            try:
                params = json.loads(data.decode("utf8"))
            except ValueError as e:
                params = ""
                self.logger.error(
                    "Message body {} does not contain valid JSON".format(data)
                )

            # ------------------------------------------------------------------

            """
            Define nested functions as handlers for each supported API action.
            Using nested functions so we can use the context from handle_post.

            That the methods are prefixed with "aws_api_" is a mitigation against
            accidentally or deliberately placing an invalid action in the API.
            """
            def aws_api_CreateStateMachine():
                """
                https://docs.aws.amazon.com/step-functions/latest/apireference/API_CreateStateMachine.html
                """
                name = params.get("name")
                if not valid_name(name):
                    self.logger.warning(
                        "RestAPI CreateStateMachine: {} is an invalid name".format(name)
                    )
                    return aws_error("InvalidName"), 400

                role_arn = params.get("roleArn")
                if not valid_role_arn(role_arn):
                    self.logger.warning(
                        "RestAPI CreateStateMachine: {} is an invalid Role ARN".format(
                            role_arn
                        )
                    )
                    return aws_error("InvalidArn"), 400

                # Form stateMachineArn from roleArn and name
                arn = parse_arn(role_arn)
                state_machine_arn = create_arn(
                    service="states",
                    region=self.region,
                    account=arn["account"],
                    resource_type="stateMachine",
                    resource=name,
                )

                # Get State Machine type (STANDARD or EXPRESS) if supplied
                type = params.get("type", "STANDARD")
                if type not in {"STANDARD", "EXPRESS"}:
                    self.logger.error(
                        "RestAPI CreateStateMachine: State Machine type {} "
                        "is not supported".format(type)
                    )
                    return aws_error("StateMachineTypeNotSupported"), 400

                """
                Look up stateMachineArn. Use get() not get_cached_view() here as
                calls to CreateStateMachine might reasonably *expect* no match.
                """
                match = self.asl_store.get(state_machine_arn)
                if match:
                    # Info seems more appropriate than error here as creation is
                    # an idempotent action.
                    self.logger.info(
                        "RestAPI CreateStateMachine: State Machine {} already exists".format(
                            state_machine_arn
                        )
                    )
                    return aws_error("StateMachineAlreadyExists"), 400

                definition = params.get("definition", "")
                """
                First check if the definition length has exceeded the 1048576
                character limit described in the CreateStateMachine API page.
                https://docs.aws.amazon.com/step-functions/latest/apireference/API_CreateStateMachine.html
                """
                if len(definition) == 0 or len(definition) > MAX_STATE_MACHINE_LENGTH:
                    self.logger.error(
                        "RestAPI CreateStateMachine: Invalid definition size for State Machine '{}'.".format(name)
                    )
                    return aws_error("InvalidDefinition"), 400

                try:
                    definition = json.loads(definition)
                except ValueError as e:
                    definition = None
                    self.logger.error(
                        "RestAPI CreateStateMachine: State Machine definition {} does not contain valid JSON".format(
                            params.get("definition")
                        )
                    )
                    return aws_error("InvalidDefinition"), 400

                if not (name and role_arn and definition):
                    self.logger.warning(
                        "RestAPI CreateStateMachine: name, roleArn and definition must be specified"
                    )
                    return aws_error("MissingRequiredParameter"), 400

                # TODO ASL Validator??

                creation_date = time.time()
                self.asl_store[state_machine_arn] = {
                    "creationDate": creation_date,
                    "definition": definition,
                    "name": name,
                    "roleArn": role_arn,
                    "stateMachineArn": state_machine_arn,
                    "updateDate": creation_date,
                    "status": "ACTIVE",
                    "type": type,
                }

                resp = {
                    "creationDate": creation_date,
                    "stateMachineArn": state_machine_arn,
                }

                return jsonify(resp), 200

            def aws_api_ListStateMachines():
                """
                https://docs.aws.amazon.com/step-functions/latest/apireference/API_ListStateMachines.html
                """
                # TODO handle nextToken stuff
                next_token = ""

                """
                Populate response using list and dict comprehensions
                https://www.pythonforbeginners.com/basics/list-comprehensions-in-python
                https://stackoverflow.com/questions/5352546/extract-subset-of-key-value-pairs-from-python-dictionary-object
                """
                state_machines = [
                    {
                        k1: v[k1] for k1 in ("creationDate", "name",
                            "stateMachineArn", "type")
                    }
                    for k, v in self.asl_store.items()
                ]

                resp = {
                    "stateMachines": state_machines
                }
                if next_token:
                    resp["nextToken"] = next_token

                return jsonify(resp), 200

            def aws_api_DescribeStateMachine():
                """
                https://docs.aws.amazon.com/step-functions/latest/apireference/API_DescribeStateMachine.html
                """
                state_machine_arn = params.get("stateMachineArn")
                if not state_machine_arn:
                    self.logger.warning(
                        "RestAPI DescribeStateMachine: stateMachineArn must be specified"
                    )
                    return aws_error("MissingRequiredParameter"), 400

                if not valid_state_machine_arn(state_machine_arn):
                    self.logger.warning(
                        "RestAPI DescribeStateMachine: {} is an invalid State Machine ARN".format(
                            state_machine_arn
                        )
                    )
                    return aws_error("InvalidArn"), 400

                """
                Look up stateMachineArn. Using get_cached_view() here means that
                the state_machine is JSON serialisable, as the cached view is a
                simple dict rather than say a RedisDict.
                """
                state_machine = self.asl_store.get_cached_view(state_machine_arn)
                if not state_machine:
                    self.logger.info(
                        "RestAPI DescribeStateMachine: State Machine {} does not exist".format(
                            state_machine_arn
                        )
                    )
                    return aws_error("StateMachineDoesNotExist"), 400
                
                """
                In the API the "definition" field is actually a string not a
                JSON object, hence the json.dumps() here. We do the conversion
                here rather than storing it as a string because the State Engine
                uses the deserialised definition as a key part of its core
                state transition behaviour.
                """
                resp = state_machine.copy()
                resp["definition"] = json.dumps(state_machine["definition"])

                return jsonify(resp), 200

            def aws_api_DescribeStateMachineForExecution():
                """
                https://docs.aws.amazon.com/step-functions/latest/apireference/API_DescribeStateMachineForExecution.html
                """
                execution_arn = params.get("executionArn")
                if not execution_arn:
                    self.logger.warning(
                        "RestAPI DescribeStateMachineForExecution: executionArn must be specified"
                    )
                    return aws_error("MissingRequiredParameter"), 400

                if not valid_execution_arn(execution_arn):
                    self.logger.warning(
                        "RestAPI DescribeStateMachineForExecution: {} is an invalid Execution ARN".format(
                            execution_arn
                        )
                    )
                    return aws_error("InvalidArn"), 400

                # Look up executionArn
                execution = self.executions.get(execution_arn)
                if not execution:
                    self.logger.info(
                        "RestAPI DescribeStateMachineForExecution: Execution {} does not exist".format(
                            execution_arn
                        )
                    )
                    return aws_error("ExecutionDoesNotExist"), 400

                state_machine_arn = execution.get("stateMachineArn")

                if not valid_state_machine_arn(state_machine_arn):
                    self.logger.warning(
                        "RestAPI DescribeStateMachineForExecution: {} is an invalid State Machine ARN".format(
                            state_machine_arn
                        )
                    )
                    return aws_error("InvalidArn"), 400

                # Look up stateMachineArn
                state_machine = self.asl_store.get_cached_view(state_machine_arn)
                if not state_machine:
                    self.logger.info(
                        "RestAPI DescribeStateMachineForExecution: State Machine {} does not exist".format(
                            state_machine_arn
                        )
                    )
                    return aws_error("StateMachineDoesNotExist"), 400

                """
                As with DescribeStateMachine the "definition" field is actually
                a string not a JSON object, hence the json.dumps() here.
                """
                resp = {
                    k: state_machine[k] for k in ("definition", "name", "roleArn",
                        "stateMachineArn", "updateDate")
                }
                resp["definition"] = json.dumps(state_machine["definition"])

                return jsonify(resp), 200

            def aws_api_UpdateStateMachine():
                """
                https://docs.aws.amazon.com/step-functions/latest/apireference/API_UpdateStateMachine.html
                """
                state_machine_arn = params.get("stateMachineArn")
                if not state_machine_arn:
                    self.logger.warning(
                        "RestAPI UpdateStateMachine: stateMachineArn must be specified"
                    )
                    return aws_error("MissingRequiredParameter"), 400

                if not valid_state_machine_arn(state_machine_arn):
                    self.logger.warning(
                        "RestAPI UpdateStateMachine: {} is an invalid State Machine ARN".format(
                            state_machine_arn
                        )
                    )
                    return aws_error("InvalidArn"), 400

                """
                Look up stateMachineArn. Use get() rather than get_cached_view()
                as we are going to be updating the retrieved State Machine.
                """
                state_machine = self.asl_store.get(state_machine_arn)
                if not state_machine:
                    self.logger.info(
                        "RestAPI UpdateStateMachine: State Machine {} does not exist".format(
                            state_machine_arn
                        )
                    )
                    return aws_error("StateMachineDoesNotExist"), 400

                role_arn = params.get("roleArn")
                if role_arn:
                    if not valid_role_arn(role_arn):
                        self.logger.warning(
                            "RestAPI UpdateStateMachine: {} is an invalid Role ARN".format(
                                role_arn
                            )
                        )
                        return aws_error("InvalidArn"), 400
                    state_machine["roleArn"] = role_arn

                definition = params.get("definition", "")
                if definition:
                    """
                    First check if the definition length has exceeded the 1048576
                    character limit described in the UpdateStateMachine API page.
                    https://docs.aws.amazon.com/step-functions/latest/apireference/API_UpdateStateMachine.html
                    """
                    if len(definition) == 0 or len(definition) > MAX_STATE_MACHINE_LENGTH:
                        self.logger.error(
                            "RestAPI CreateStateMachine: Invalid definition size for State Machine '{}'.".format(name)
                        )
                        return aws_error("InvalidDefinition"), 400

                    try:
                        definition = json.loads(definition)
                    except ValueError as e:
                        definition = None
                        self.logger.error(
                            "RestAPI UpdateStateMachine: State Machine definition {} does not contain valid JSON".format(
                                params.get("definition")
                            )
                        )
                        return aws_error("InvalidDefinition"), 400

                    # TODO ASL Validator??
                    state_machine["definition"] = definition

                if not role_arn and not definition:
                    self.logger.warning(
                        "RestAPI UpdateStateMachine: either roleArn or definition must be specified"
                    )
                    return aws_error("MissingRequiredParameter"), 400

                update_date = time.time()
                state_machine["updateDate"] = update_date

                self.asl_store[state_machine_arn] = state_machine

                resp = {"updateDate": update_date}

                return jsonify(resp), 200

            def aws_api_DeleteStateMachine():
                """
                https://docs.aws.amazon.com/step-functions/latest/apireference/API_DeleteStateMachine.html
                TODO This should really mark the state machine for deletion and
                "The state machine itself is deleted after all executions are 
                completed or deleted."
                """
                state_machine_arn = params.get("stateMachineArn")
                if not state_machine_arn:
                    self.logger.warning(
                        "RestAPI DeleteStateMachine: stateMachineArn must be specified"
                    )
                    return aws_error("MissingRequiredParameter"), 400

                if not valid_state_machine_arn(state_machine_arn):
                    self.logger.warning(
                        "RestAPI DeleteStateMachine: {} is an invalid State Machine ARN".format(
                            state_machine_arn
                        )
                    )
                    return aws_error("InvalidArn"), 400

                # Look up stateMachineArn
                state_machine = self.asl_store.get_cached_view(state_machine_arn)
                if not state_machine:
                    self.logger.info(
                        "RestAPI DeleteStateMachine: State Machine {} does not exist".format(
                            state_machine_arn
                        )
                    )
                    return aws_error("StateMachineDoesNotExist"), 400

                del self.asl_store[state_machine_arn]

                return "", 200

            def aws_api_StartExecution():
                """
                https://docs.aws.amazon.com/step-functions/latest/apireference/API_StartExecution.html
                """
                # print(params)
                state_machine_arn = params.get("stateMachineArn")
                if not state_machine_arn:
                    self.logger.warning(
                        "RestAPI StartExecution: stateMachineArn must be specified"
                    )
                    return aws_error("MissingRequiredParameter"), 400

                if not valid_state_machine_arn(state_machine_arn):
                    self.logger.warning(
                        "RestAPI StartExecution: {} is an invalid State Machine ARN".format(
                            state_machine_arn
                        )
                    )
                    return aws_error("InvalidArn"), 400

                """
                If name isn't provided create one from a UUID. TODO names should
                be unique within a 90 day period, at the moment there is no code
                to check for uniqueness of provided names so client code that
                doesn't honour this may currently succeed in this implementation
                but fail if calling real AWS StepFunctions.
                """
                name = params.get("name", str(uuid.uuid4()))
                if not valid_name(name):
                    self.logger.warning(
                        "RestAPI StartExecution: {} is an invalid name".format(name)
                    )
                    return aws_error("InvalidName"), 400

                input = params.get("input", "{}")
                """
                First check if the input length has exceeded the 262144 character
                quota described in Stepfunction Quotas page.
                https://docs.aws.amazon.com/step-functions/latest/dg/limits.html
                """
                if len(input) > MAX_DATA_LENGTH:
                    self.logger.error(
                        "RestAPI StartExecution: input size for execution '{}' exceeds "
                        "the maximum number of characters service limit.".format(name)
                    )
                    return aws_error("InvalidExecutionInput"), 400

                try:
                    input = json.loads(input)
                except TypeError as e:
                    self.logger.error("RestAPI StartExecution: Invalid input, {}".format(e))
                    return aws_error("InvalidExecutionInput"), 400
                except ValueError as e:
                    self.logger.error(
                        "RestAPI StartExecution: input {} does not contain valid JSON".format(
                            input
                        )
                    )
                    return aws_error("InvalidExecutionInput"), 400

                # Look up stateMachineArn
                state_machine = self.asl_store.get_cached_view(state_machine_arn)
                if not state_machine:
                    self.logger.info(
                        "RestAPI StartExecution: State Machine {} does not exist".format(
                            state_machine_arn
                        )
                    )
                    return aws_error("StateMachineDoesNotExist"), 400


                # Form executionArn from stateMachineArn and name
                arn = parse_arn(state_machine_arn)
                execution_arn = create_arn(
                    service="states",
                    region=arn.get("region", self.region),
                    account=arn["account"],
                    resource_type="execution",
                    resource=arn["resource"] + ":" + name,
                )

                with opentracing.tracer.start_active_span(
                    operation_name="StartExecution:ExecutionLaunching",
                    child_of=span_context("http_headers", request.headers, self.logger),
                    tags={
                        "component": "rest_api",
                        "execution_arn": execution_arn
                    }
                ) as scope:
                    """
                    The application context is described in the AWS documentation:
                    https://docs.aws.amazon.com/step-functions/latest/dg/input-output-contextobject.html
                    """
                    # https://stackoverflow.com/questions/8556398/generate-rfc-3339-timestamp-in-python
                    start_time = datetime.now(timezone.utc).astimezone().isoformat()
                    context = {
                        "Tracer": inject_span("text_map", scope.span, self.logger),
                        "Execution": {
                            "Id": execution_arn,
                            "Input": input,
                            "Name": name,
                            "RoleArn": state_machine.get("roleArn"),
                            "StartTime": start_time,
                        },
                        "State": {"EnteredTime": start_time, "Name": ""},  # Start state
                        "StateMachine": {
                            "Id": state_machine_arn,
                            "Name": state_machine.get("name"),
                        },
                    }

                    event = {"data": input, "context": context}

                    """
                    threadsafe=True is important here as the RestAPI runs in a
                    different thread to the main event_dispatcher loop.
                    """
                    try:
                        self.event_dispatcher.publish(
                            event, threadsafe=True, start_execution=True
                        )
                    except:
                        message = ("RestAPI StartExecution: Internal messaging "
                                  "error, start message could not be published.")
                        self.logger.error(message)
                        return aws_error("InternalError", message), 500

                    resp = {"executionArn": execution_arn, "startDate": time.time()}

                    return jsonify(resp), 200

            def aws_api_ListExecutions():
                """
                https://docs.aws.amazon.com/step-functions/latest/apireference/API_ListExecutions.html
                """
                state_machine_arn = params.get("stateMachineArn")
                if not state_machine_arn:
                    self.logger.warning(
                        "RestAPI ListExecutions: stateMachineArn must be specified"
                    )
                    return aws_error("MissingRequiredParameter"), 400

                if not valid_state_machine_arn(state_machine_arn):
                    self.logger.warning(
                        "RestAPI ListExecutions: {} is an invalid State Machine ARN".format(
                            state_machine_arn
                        )
                    )
                    return aws_error("InvalidArn"), 400

                # Look up stateMachineArn
                state_machine = self.asl_store.get_cached_view(state_machine_arn)
                if not state_machine:
                    self.logger.info(
                        "RestAPI ListExecutions: State Machine {} does not exist".format(
                            state_machine_arn
                        )
                    )
                    return aws_error("StateMachineDoesNotExist"), 400

                status_filter = params.get("statusFilter")
                if status_filter and status_filter not in {
                    "RUNNING",
                    "SUCCEEDED",
                    "FAILED",
                    "TIMED_OUT",
                    "ABORTED",
                }:
                    status_filter = None

                """
                Populate response using list and dict comprehensions
                https://www.pythonforbeginners.com/basics/list-comprehensions-in-python
                https://stackoverflow.com/questions/5352546/extract-subset-of-key-value-pairs-from-python-dictionary-object

                TODO handle nextToken stuff.
                Note that ListExecutions is potentially a very expensive call as
                there might well be a large number of executions for any given
                State Machine and moreover the execution details are stored as
                Redis hashes that are themselves keyed by the execution ARN. In
                other words it is not *natively* a list and under the covers
                listing the executions is implemented by a redis.scan. One
                option for improving things might be to use the next_token to
                wrap a scan cursor. That approach should works as the maxResults
                in the API call is only a hint and the actual number of results
                returned per call might be fewer than the specified maximum, so
                that fits somewhat to the constraints of Redis scan cursors.
                """
                next_token = ""

                executions = [
                    {
                        k1: v[k1] for k1 in ("executionArn", "name", "startDate",       
                            "stateMachineArn", "status", "stopDate")
                    }
                    for k, v in self.executions.items()
                    if v["stateMachineArn"] == state_machine_arn
                    and (status_filter == None or v["status"] == status_filter)
                ]

                resp = {
                    "executions": executions
                }
                if next_token:
                    resp["nextToken"] = next_token

                return jsonify(resp), 200

            def aws_api_DescribeExecution():
                """
                https://docs.aws.amazon.com/step-functions/latest/apireference/API_DescribeExecution.html
                """
                execution_arn = params.get("executionArn")
                if not execution_arn:
                    self.logger.warning(
                        "RestAPI DescribeExecution: executionArn must be specified"
                    )
                    return aws_error("MissingRequiredParameter"), 400

                if not valid_execution_arn(execution_arn):
                    self.logger.warning(
                        "RestAPI DescribeExecution: {} is an invalid Execution ARN".format(
                            execution_arn
                        )
                    )
                    return aws_error("InvalidArn"), 400

                # Look up executionArn
                execution = self.executions.get(execution_arn)
                if not execution:
                    self.logger.info(
                        "RestAPI DescribeExecution: Execution {} does not exist".format(
                            execution_arn
                        )
                    )
                    return aws_error("ExecutionDoesNotExist"), 400

                if not isinstance(execution , dict):  # May be (non JSON) RedisDict
                    execution = dict(execution)
                return jsonify(execution), 200

            def aws_api_GetExecutionHistory():
                """
                https://docs.aws.amazon.com/step-functions/latest/apireference/API_GetExecutionHistory.html
                """
                # print(params)

                execution_arn = params.get("executionArn")
                if not execution_arn:
                    self.logger.warning(
                        "RestAPI GetExecutionHistory: executionArn must be specified"
                    )
                    return aws_error("MissingRequiredParameter"), 400

                if not valid_execution_arn(execution_arn):
                    self.logger.warning(
                        "RestAPI GetExecutionHistory: {} is an invalid Execution ARN".format(
                            execution_arn
                        )
                    )
                    return aws_error("InvalidArn"), 400

                reverse_order = params.get("reverseOrder", False)

                # Look up executionArn
                history = self.execution_history.get(execution_arn)
                if not history:
                    self.logger.info(
                        "RestAPI GetExecutionHistory: Execution {} does not exist".format(
                            execution_arn
                        )
                    )
                    return aws_error("ExecutionDoesNotExist"), 400

                """
                Reverse via slicing: [start:stop:step] so step is -1
                https://stackoverflow.com/questions/3940128/how-can-i-reverse-a-list-in-python

                TODO handle nextToken stuff.

                Note that GetExecutionHistory is potentially an expensive call
                if the history is large. The store self.execution_history has
                list semantics, but is backed by an external (e.g. Redis) store.
                Under the covers it will do a redis.lrange, so the next_token
                behaviour when implemented should "slice" the appropriate range.
                Note that doing this for GetExecutionHistory should be easier
                than for ListExecutions - see comment in ListExecutions for why.
                """
                if reverse_order:
                    history = history[::-1]
                else:
                    history = history[:]

                next_token = ""

                resp = {"events": history}
                if next_token:
                    resp["nextToken"] = next_token

                return jsonify(resp), 200

            def aws_api_InvalidAction():
                self.logger.error("RestAPI invalid action: {}".format(action))
                return "InvalidAction", 400

            # ------------------------------------------------------------------

            # Use the API action to dynamically invoke the appropriate handler.
            try:
                value, code = locals().get("aws_api_" + action, aws_api_InvalidAction)()
                return value, code
            except Exception as e:
                self.logger.error(
                    "RestAPI action {} failed unexpectedly with exception: {}".format(
                        action, e
                    )
                )
                return "InternalError", 500

        # Have an endpoint to check the health of the ASL engine
        @app.route("/health")
        def health_check():
            # Check if the session channel is open (alike session.channel.is_open())
            # Give response dependent on outcome of session
            if self.event_dispatcher.session.is_open():
                return "Ok", 200
            else:
                return "Service Unavailable: Service liveness probe failed due to an internal error.", 503

        return app


from asyncio.log import logger
from unittest import result
from asl_workflow_engine.logger import init_logging
from asl_workflow_engine.state_engine import StateEngine
from threading import Timer
from asl_workflow_engine.arn import parse_arn
import json
import unittest
import random,string,os
import shutil
import re
ASL = """{
  "Comment": "A Hello World example of the Amazon States Language using Pass states",
  "StartAt": "SplitVideo",
  "States": {
    "SplitVideo": {
      "Type": "Task",
      "Resource": "arn:aws:lambda:us-east-1:187983511897:function:SplitVideo",
      "OutputPath":"$$.Execution",
      "ResultPath":"$",
      "Parameters": {
        "Input.$": "$",
        "WorkDir":"test_video_trans_func/SplitVideo"
      },
      "Next": "SegmentalVideoSolution"
    },
    "SegmentalVideoSolution": {
      "Type": "Map",
      "InputPath":"$",
      "ItemsPath": "$.Input.item",
      "ResultPath": "$",
      "Iterator": {
        "StartAt": "VideoSolution",
        "States": {
          "VideoSolution": {
            "Type": "Task",
            "Resource": "arn:aws:lambda:us-east-1:187983511897:function:VideoSolution",
            "Parameters":{
                "WorkDir":"test_video_trans_func/VideoSolution",
                "Dependencies":["SplitVideo"]
            },
            "End": true
          }
        }
      },
      "Next": "ComposeVideo"
    },
    "ComposeVideo": {
      "Type": "Task",
      "InputPath": "$",
      "OutputPath":"$",
      "Resource": "arn:aws:lambda:us-east-1:187983511897:function:ComposeVideo",
      "Parameters":{
        "WorkDir":"test_video_trans_func/ComposeVideo",
        "Dependencies":["VideoSolution"]
      },
      "End": true
    }
  }
}"""
"""
The application context is described in the AWS documentation:
https://docs.aws.amazon.com/step-functions/latest/dg/input-output-contextobject.html 

{
    "Execution": {
        "Id": <String>,
        "Input": <Object>,
        "StartTime": <String Format: ISO 8601>
    },
    "State": {
        "EnteredTime": <String Format: ISO 8601>,
        "Name": <String>,
        "RetryCount": <Number>
    },
    "StateMachine": {
        "Id": <String>,
        "Definition": <Object representing ASL state machine>
    },
    "Task": {
        "Token": <String>
    }
}

The most important paths for state traversal are:
$$.State.Name = the current state
$$.StateMachine.Definition = (optional) contains the complete ASL state machine
$$.StateMachine.Id = a unique reference to an ASL state machine
"""

context = '{"StateMachine": {"Id": "arn:aws:states:local:0123456789:stateMachine:simple_state_machine", "Definition": ' + str(ASL) + '}}'
output_file_key=[]

config = {"workdir":"/test_video_trans_func",
          "preload_data_key":"video.mp4",
          "preload_data_path":"test_video_trans_func/RawData/video.mp4"
          }
state_engine_config = {
            "state_engine": {
                "store_url": "ASL_store.json", 
                "execution_ttl": 500
            }
      }
class Params():
    """Class that loads hyperparameters from a json file.
        Example:
        ```
        params = Params(json_path)
        print(params.learning_rate)
        params.learning_rate = 0.5  # change the value of learning_rate in params
        ```
        """
    def __init__(self, json_path):
        with open(json_path) as f:
            params = json.load(f) 
            self.__dict__.update(params)

    def save(self, json_path):
        with open(json_path, 'w') as f:
            json.dump(self.__dict__, f, indent=4)  

    def update(self, json_path):
        """Loads parameters from json file"""
        with open(json_path) as f:
            params = json.load(f)
            self.__dict__.update(params)

    @property
    def dict(self):
        """Gives dict-like access to Params instance by `params.dict['learning_rate']"""
        return self.__dict__

class EventDispatcherStub(object):

    def __init__(self, state_engine, config):

        """
        Create an association with the state engine and give that a reference
        back to this event dispatcher so that it can publish events and make
        use of the set_timeout time scheduler.
        """
        self.state_engine = state_engine
        self.state_engine.event_dispatcher = self
        self.message_count = -1

    """
    This simple threaded timeout should work OK, the real timeout is actually
    implemented using Pika's connection.call_later() which is single threaded
    and handled within Pika's event loop. That approach plays much better with
    Pika's event loop.
    """
    def set_timeout(self, callback, delay):
        t = Timer(delay/1000, callback)
        t.start()
        return t

    def dispatch(self, message):
        """
        Start at -1 and increment before call to notify as this stub, unlike
        the real EventDispatcher will recursively call dispatch as the State
        Engine calls publish, this means code after the call to notify won't
        be reached when one might expect it to.
        """
        self.message_count += 1
        # The notify method expects a JSON object not a string.
        self.state_engine.notify(json.loads(message), self.message_count)

    def acknowledge(self, id):
        pass

    def publish(self, item):
        # Save Execution StartTime to use when validating output
        if self.message_count == 0:
            self.execution_start_time = item["context"]["Execution"]["StartTime"]
        # Convert event back to JSON string for dispatching.
        self.dispatch(json.dumps(item))

    def broadcast(self, subject, message, carrier_properties=None):
        self.output_event = message

def genRandomString(slen=10):
    return ''.join(random.sample(string.ascii_letters + string.digits, slen))

def get_lower_case_name(text):
    lst = []
    for index, char in enumerate(text):
        if char.isupper() and index != 0:
            lst.append("-")
        lst.append(char)

    return "".join(lst).lower()

def execute_task_stub(resource_arn, parameters, callback,timeout, context):
    
    arn = parse_arn(resource_arn)
    service = arn["service"]  # e.g. rpcmessage, fn, openfaas, lambda
    resource_type = arn["resource_type"]  # Should be function most times
    resource = arn["resource"]  # function-name
    configDir= parameters["WorkDir"]+'/config.json'  
    config_context={
            "resource":resource,
            "service":service,
            "resource_type":resource_type,
            "function_url":"192.168.122.201:31112/function/"+get_lower_case_name(resource)
          }
    if resource=='SplitVideo':
        config_context['dependencies_output_key']=['video.mp4']
    with open(configDir,'w+') as config:
        config.write(json.dumps(config_context))      
    result = {"parameters":parameters,"context":context}
    print(result)     
    callback(result)

def execute_task_openfaas(resource_arn, parameters, callback,timeout, context):
    arn = parse_arn(resource_arn)
    resource = arn["resource"]
    workdir=parameters['WorkDir']
    result = {"parameters":parameters,"Output":context}
    func_dir="test_video_trans_func/"+resource+"/"
    print(func_dir)
    mycopyfile("test_video_trans_func/handler_templete.py",func_dir,"handler.py")
    callback(result)

def execute_openfaas_request(resource_arn, parameters, callback,timeout, context):
    arn = parse_arn(resource_arn)
    resource = arn["resource"]  # function-name
    function_url="192.168.122.201:31112/function/" +get_lower_case_name(resource)
    if "Dependencies" in parameters:
        dependency_key_list=[]
        for i in parameters['Dependencies']:
            with open('test_video_trans_func/'+i+"/config.json","r") as f:
                dependency_func_config=json.loads(f.read())
                try:
                    for i in dependency_func_config['output_key_list']:
                        dependency_key_list.append(i)
                except:
                    print('-------------------error----------------------')
        with open('test_video_trans_func/'+resource+"/config.json","w") as f:
            config={'resource':resource,'arn':arn,"function_url":function_url,'dependency_key_list':dependency_key_list}
            f.write(json.dumps(config))
            if 'Branch' in context['State']:
                index=context['State']['Branch'][0]['Index']
                parameters['dependency_key_list']=[dependency_key_list[0][index]]
                parameters['index']=str(index)
            else:
                parameters['dependency_key_list']=[]
                for i in dependency_key_list:
                    parameters['dependency_key_list'].append(i[0])
    cmd= ''' curl '''+function_url+''' --data '''+"'"+json.dumps(parameters)+"'"
    print("cmd***********************************",cmd)
    result=os.popen(cmd)
    cmd_context=result.read()
    print(cmd_context)
    cmd_list=re.split('\n',cmd_context)
    cmd_return=cmd_list[len(cmd_list)-2]
    cmd_return=cmd_return.replace("'", '"')
    cmd_obj=json.loads(cmd_return)
    function_name = get_lower_case_name(resource)
    output_key_list=cmd_obj['output_key_list']
    with open('test_video_trans_func/'+resource+"/config.json","r+") as f:
        config=json.loads(f.read())
    if('output_key_list' in config and len(config['output_key_list'])>0):
        config['output_key_list'].append(output_key_list)
    else:
        config['output_key_list']=[]
        config['output_key_list'].append(output_key_list)
    with open('test_video_trans_func/'+resource+"/config.json","w+") as f:
        f.write(json.dumps(config))
    result=cmd_obj
    callback(result)
     
def mycopyfile(srcfile,dstpath,fname):                       
    if not os.path.isfile(srcfile):
        print ("%s not exist!"%(srcfile))
    else:
        if not os.path.exists(dstpath):
            os.makedirs(dstpath)                                                  
        shutil.copy(srcfile, dstpath + fname)          
        # print ("copy %s -> %s"%(srcfile, dstpath + fname))

class TestVideoTrans(unittest.TestCase):

    def setUp(self):
      # Initialise logger
        logger=init_logging(log_name="function_build_engine")

    def test_get_folder(self):
        state_engine = StateEngine(state_engine_config)
        state_engine.task_dispatcher.execute_task = execute_task_stub
        self.event_dispatcher = EventDispatcherStub(state_engine, config)
        self.event_dispatcher.dispatch('{"data":{"item":[1,2,3]}, "context": ' + context + '}')

    def test_execute_task_openfaas(self):
        state_engine = StateEngine(state_engine_config)
        # Stub out the real TaskDispatcher execute_task
        state_engine.task_dispatcher.execute_task = execute_task_openfaas
        self.event_dispatcher = EventDispatcherStub(state_engine, config)
        self.event_dispatcher.dispatch('{"data":{"item":[1,2,3]}, "context": ' + context + '}')

    def test_async_api(self):
        cmd='cd test_video_trans_func && faas up -f stack.yaml'
        os.system(cmd)
        state_engine = StateEngine(state_engine_config)
        self.event_dispatcher = EventDispatcherStub(state_engine, config)
        state_engine.task_dispatcher.execute_task=execute_openfaas_request
        self.event_dispatcher.dispatch('{"data":{"item":[1,2,3]}, "context": ' + context + '}')
        

if __name__ == '__main__':
    unittest.main()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Run with:
# PYTHONPATH=.. python3 map2.py
#
"""
This test uses the AWS python SDK boto3 to access our local ASL Workflow Engine
https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/stepfunctions.html
https://docs.aws.amazon.com/code-samples/latest/catalog/code-catalog-python-example_code-stepfunctions.html

This example is a minimal example to illustrate use of the Map State
MaxConcurrency field, which can help limit throughput and somewhat serialise
access to expensive (or otherwise limited) resources to avoid overwhelming them
https://states-language.net/spec.html#map-state
"""

import sys
assert sys.version_info >= (3, 0) # Bomb out if not running Python3

import boto3, time
from botocore.exceptions import ClientError

from asl_workflow_engine.logger import init_logging
from asl_workflow_engine.open_tracing_factory import create_tracer

ASL = """{
  "Comment": "Map Example.",
  "StartAt": "Validate-All",
  "States": {
    "Validate-All": {
      "Type": "Map",
      "InputPath": "$",
      "ItemsPath": "$.items",
      "MaxConcurrency": 3,
      "ResultPath": "$.items",
      "Parameters": {
        "item.$": "$$.Map.Item.Value",
        "index.$": "$$.Map.Item.Index"
      },
      "Iterator": {
        "StartAt": "Validate",
        "States": {
          "Validate": {
            "Type": "Pass",
            "Comment": "The Result from each iteration should be its *effective* input, which is a JSON node that contains both the current item data from the context object, and the courier information from the 'delivery-partner' field from the Map state input.",
            "End": true
          }
        }
      },
      "End": true
    }
  }
}"""

items = ["""
{
    "items": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]
}
"""]

if __name__ == '__main__':
    # Initialise logger
    logger = init_logging(log_name="map2")

    # Initialising OpenTracing. It's important to do this before the boto3.client
    # call as create_tracer "patches" boto3 to add the OpenTracing hooks.
    create_tracer("map2", {"implementation": "Jaeger"})

    # Initialise the boto3 client setting the endpoint_url to our local
    # ASL Workflow Engine
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html#boto3.session.Session.client
    sfn = boto3.client("stepfunctions", endpoint_url="http://localhost:4584")
    state_machine_arn = "arn:aws:states:local:0123456789:stateMachine:map2"

    def create_state_machines():
        # Create state machine using a dummy roleArn. If it already exists an
        # exception will be thrown, we ignore that but raise other exceptions.
        try:
            response = sfn.create_state_machine(
                name="map2", definition=ASL,
                roleArn="arn:aws:iam::0123456789:role/service-role/MyRole"
            )
        except sfn.exceptions.StateMachineAlreadyExists as e:
            #print(e.response)

            response = sfn.update_state_machine(
                stateMachineArn=state_machine_arn,
                definition=ASL
            )
        except ClientError as e:
            logger.error(e)


    create_state_machines()

    # Loop through items invoking a new state machine execution for each item
    for item in items:
        try:
            response = sfn.start_execution(
                stateMachineArn=state_machine_arn,
                #name=EXECUTION_NAME, # If not specified a UUID is assigned
                input=item
            )
        except sfn.exceptions.StateMachineDoesNotExist as e:
            logger.info(e)

            create_state_machines()
            response = sfn.start_execution(
                stateMachineArn=state_machine_arn,
                #name=EXECUTION_NAME, # If not specified a UUID is assigned
                input=item
            )
        except ClientError as e:
            logger.error(e)

    time.sleep(1)  # Give OpenTracing a chance to flush buffer before exiting




#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""
https://states-language.net/spec.html#filters

A state may want to process only a subset of its input data, and may want that
data structured differently from the way it appears in the input. Similarly, it
may want to control the format and content of the data that it passes on as
output.

Fields named “InputPath”, “Parameters”, “OutputPath”, and “ResultPath” exist to
support this. Any state except for a Fail State MAY have “InputPath” and
“OutputPath”. States which potentially generate results MAY have “ResultPath”
and “Parameters”: Pass State, Task State, and Parallel State.
"""

import sys
assert sys.version_info >= (3, 0)  # Bomb out if not running Python3

import re

"""
ASL paths use JSONPath.
https://goessner.net/articles/JsonPath/
http://www.ultimate.com/phil/python/#jsonpath
Tested using jsonpath 0.82. Note jsponpath_rw was tried but doesn't seem to
correctly support many of the test cases from the goessner link above.
"""
from jsonpath import jsonpath  # pip3 install jsonpath

from asl_workflow_engine.asl_exceptions import *

try:  # Attempt to use ujson if available https://pypi.org/project/ujson/
    import ujson as json
except:  # Fall back to standard library json
    import json


#def apply_jsonpath(input, path="$", throw_exception_on_failed_match=False):
def apply_jsonpath(input, path="$", throw_exception_on_failed_match=True):
    """
    Performs the InputPath and OutputPath logic described in the ASL spec.
    https://states-language.net/spec.html#filters
    This is mostly just calling jsonpath() and applying the specified defaults.

    The throw_exception_on_failed_match parameter allows callers to select
    whether a failed JSONPath match will throw an exception or just return {}.
    This is because it's a little unclear what the most useful behaviour is
    in this circumstance. The default was originally to return an empty object
    this is because the behaviour of InputPath and OutputPath being null is to
    return an empty JSON object and a null JSONPath result is consistent with
    this. OTOH throwing an exception on a match failure is also useful....
    AWS StepFunctions appears to fail executions in these cases, but it's a
    bit poorly specified and largely depends on the underlying JSONPath engine. 
    What to do in this case isn't obvious from the JSONPath specification either.
    https://goessner.net/articles/JsonPath/
    https://www.tbray.org/ongoing/When/201x/2017/04/14/JsonPath-Needs-Work
    """
    if input == None or path == None:
        return {}
    if path == "$":
        return input
    result = jsonpath(input, path)

    if result == False:
        if throw_exception_on_failed_match:
            raise PathMatchFailure(
                "Invalid path '{}' applied to input '{}'".format(path, input)
            )
        else:
            return {}

    """
    The following is a little subtle. Unfortunately the JSONPath specification
    is vague on a few points and some implementations, such as Python jsonpath,
    return a list of matches, but for most scenarios if a single item matches
    it is more intuitive to have that item returned rather than a list that
    contains that item. Other implementations such as the Java Jayway (which
    I think is the one used in AWS Step Functions) behave in that way. An
    exception is where the path contains an array slice operator because then
    we intuitively expect to return an array/list even if only a single item
    is matched. The code below attempts to handle that array slice edge case.
    """
    if len(result) == 1:
        path_has_slice = re.search(r"\[.*:.*\]", path)
        if not path_has_slice:
            return result[0]

    return result

#def apply_path(input, context, path="$", throw_exception_on_failed_match=False):
def apply_path(input, context, path="$", throw_exception_on_failed_match=True):
    """
    https://states-language.net/spec.html#path

    A Path is a string, beginning with "$", used to identify components with a
    JSON text. The syntax is that of JSONPath.

    When a Path begins with "$$", two dollar signs, this signals that it is
    intended to identify content within the Context Object. The first dollar
    sign is stripped, and the remaining text, which begins with a dollar sign,
    is interpreted as the JSONPath applying to the Context Object.
    """
    if path == None or not isinstance(path, str):
        return {}
    if not path.startswith("$"):
        raise ParameterPathFailure("{} must be a JSONPath".format(path))
    if path.startswith("$$"):  # Use Context object, not input
        path = path[1:]  # Strip leading "$" from context path
        return apply_jsonpath(context, path, throw_exception_on_failed_match)
    else:
        return apply_jsonpath(input, path, throw_exception_on_failed_match)

def get_full_jsonpath(input, path):
    """
    Given an input and a JSONPath query path return the full JSONPath path of
    the query. This can be useful because JSONPath doesn't support getting
    parents of matching nodes
    """
    result = jsonpath(input, path, result_type="PATH")
    """
    #if result and len(result) == 1:
    TODO if len(result) > 1 then it means multiple states with the same name,
    which is an invalid state machine. Need to work out the best place to check
    for that situation, which is probably not here. For now always return the
    first matching result.
    """
    if result:
        return result[0]
    else:
        return ""

def apply_resultpath(input, result, path="$"):
    """
    Performs the ResultPath logic described in the ASL spec.
    https://states-language.net/spec.html#filters

    The value of “ResultPath” MUST be a Reference Path, which specifies the raw
    input’s combination with or replacement by the state’s result.

    The ResultPath field’s value is a Reference Path that specifies where to
    place the result, relative to the raw input. If the input has a field which
    matches the ResultPath value, then in the output, that field is discarded
    and overwritten by the state output. Otherwise, a new field is created in
    the state output.

    If the value of of ResultPath is null, that means that the state’s own raw
    output is discarded and its raw input becomes its result.
    """
    def update_path(target, keys, default):
        if len(keys) == 0:
            return default
        key = keys.pop(0)
        if isinstance(target, list):
            try:
                i = int(key)
                target[i] = update_path(target[i], keys, default)
            except (ValueError, IndexError) as e:
                raise ResultPathMatchFailure(e)
        elif isinstance(target, dict):
            try:  # Test if key is (incorrectly) an int.
                int(key)
                raise ResultPathMatchFailure(
                    "Object index {} is not a valid key string".format(key)
                )
            except ValueError:
                target[key] = update_path(target.get(key, {}), keys, default)
        else:
            raise ResultPathMatchFailure(
                "Cannot use key {} to index a primitive type".format(key)
            )
        return target

    if input == None:
        input = {}
    if path == None:
        return input
    if path == "$":
        return result
    if path.startswith("$$"):
        """
        The value of "ResultPath" MUST NOT begin with "$$"; i.e. it may not be
        used to insert content into the Context Object.
        """
        raise ResultPathMatchFailure(
            "The value of \"ResultPath\" MUST NOT begin with \"$$\""
        )

    matches = re.findall(r"[^$.[\]]+", path)  # Regex to split the reference paths
    return update_path(input, matches, result)

def evaluate_payload_template(input, context, template):
    """
    https://states-language.net/spec.html#payload-template

    A state machine interpreter dispatches data as input to tasks to do useful
    work, and receives output back from them. It is frequently desired to
    reshape input data to meet the format expectations of tasks, and similarly
    to reshape the output coming back. A JSON object structure called a Payload
    Template is provided for this purpose.

    The value of "Parameters" MUST be a Payload Template which is a JSON object,
    whose input is the result of applying the InputPath to the raw input. If the
    "Parameters" field is provided, its payload, after the extraction and
    embedding, becomes the effective input.

    The value of "ResultSelector" MUST be a Payload Template, whose input is the
    result, and whose payload replaces and becomes the effective result.

    Values from the Payload Template’s input and the Context Object can be
    inserted into the payload with a combination of a field-naming convention,
    Paths and Intrinsic Functions.

    If any field within the Payload Template (however deeply nested) has a name
    ending with the characters ".$", its value is transformed and the field is
    renamed to strip the ".$" suffix.

    If the field value begins with only one "$", the value MUST be a Path. In
    this case, the Path is applied to the Payload Template’s input and is the
    new field value.

    If the field value begins with "$$", the first dollar sign is stripped and
    the remainder MUST be a Path. In this case, the Path is applied to the
    Context Object and is the new field value.

    If the field value does not begin with "$", it MUST be an Intrinsic Function.
    The interpreter invokes the Intrinsic Function and the result is the new value.

    If the path is legal but cannot be applied successfully, the interpreter
    fails the machine execution with an Error Name of "States.ParameterPathFailure".
    If the Intrinsic Function fails during evaluation, the interpreter fails the
    machine execution with an Error Name of "States.IntrinsicFailure".

    When a field name ends with “.$” and its value can be used to generate an
    Extracted Value as described above, the field is replaced within the
    Parameters value by another field whose name is the original name minus the
    “.$” suffix, and whose value is the Extracted Value.

    This implementation extends the ASL specification a little as it also
    supports replacement of JSON array values such that for [0, 1, "$.map.c.$"]
    the third item would be replaced by the result of applying the JSONpath of
    $.map.c to the input.
    """

    def evaluate_intrinsic_function(intrinsic):

        def asl_intrinsic_Format(args):
            template_string = args[0]
            args = args[1:]
            try:
                return template_string.format(*args)
            except Exception as e:
                raise IntrinsicFailure(
                    "States.Format failed with {}.".format(e)
                )

        def asl_intrinsic_StringToJson(args):
            if len(args) != 1:
                raise IntrinsicFailure(
                    "States.StringToJson failed with too many arguments."
                )
            try:
                return json.loads(args[0])
            except Exception as e:
                raise IntrinsicFailure(
                    "States.StringToJson failed with {}.".format(e)
                )

        def asl_intrinsic_JsonToString(args):
            if len(args) != 1:
                raise IntrinsicFailure(
                    "States.JsonToString failed with too many arguments."
                )
            try:
                return json.dumps(args[0])
            except Exception as e:
                raise IntrinsicFailure(
                    "States.JsonToString failed with {}.".format(e)
                )

        def asl_intrinsic_Array(args):
            return args

        def asl_intrinsic_Default(args):
            raise IntrinsicFailure(
                "Intrinsic Function {} is not supported.".format(func)
            )


        # Extract intrinsic name and normalise it to asl_intrinsic_<name>
        func, args = intrinsic.split("(", 1)
        func = func.strip()
        normalised_func = func.replace("States.", "asl_intrinsic_")
        # Extract raw args string
        args = args.rsplit(")", 1)[0]

        """
        Extract the individual args from the raw string into a list. Intrinsic
        Function arguments may be strings enclosed by apostrophe (') characters,
        numbers, null, Paths, or nested Intrinsic Functions. The regex finds
        each valid argument as follows:
        \'.*?(?<!\\\)\'         extracts apostrophe delimited string. This uses
            a negative lookbehind to match a closing ' only if not preceeded
            by a \ in order to support escaped apostrophes in the string.
        States.*?\)             extracts nested intrinsic

        String and nested intrinsics can contain commas so we explicitly match
        those cases, but the last part of the regex '|[^\s*,]+' just matches
        anything except whitespace comma. We actually *want* a fairly loose
        match here so if we have an invalid number like f123.45 it would match
        but subsequent evaluation would raise an IntrinsicFailure which we want.
        """
        arglist = re.findall('\'.*?(?<!\\\)\'|States.*?\)|[^\s*,]+', args)

        # Evaluate the arguments
        for i, arg in enumerate(arglist):
            if arg.startswith("'"):  # It's an apostrophe delimited string
                arglist[i] = arg.strip("'")
            elif arg.startswith("$"):  # It's a path
                arglist[i] = apply_path(input, context, arg)
            elif arg.startswith("States."):  # It's a nested intrinsic function
                arglist[i] = evaluate_intrinsic_function(arg)
            elif arg == "null":
                arglist[i] = None
            elif arg == "true":
                """
                Note that the ASL spec doesn't explicitly include booleans in
                the supported Intrinsic Function arguments, however as there
                is a States.Array intrinsic that returns a JSON array containing
                the Values of the arguments, the implication is that arguments
                could be JSON primitives. It's a little unclear.
                """
                arglist[i] = True
            elif arg == "false":
                arglist[i] = False
            else:
                try:
                    arglist[i] = int(arg)
                except ValueError:
                    try:
                        arglist[i] = float(arg)
                    except ValueError:
                        raise IntrinsicFailure(
                            "Intrinsic Function {}, Invalid argument {}.".format(func, arg)
                        )

        return locals().get(
            normalised_func,
            asl_intrinsic_Default,
        )(arglist)

    def evaluate(k, v=None):
        """
        Evaluate and expand fields whose name ends with “.$” as described above
        """
        if isinstance(k, str) and k.endswith(".$"):
            k = k[:-2]  # strip ".$" from end
            v_is_path_or_intrinsic = True
        else:
            v_is_path_or_intrinsic = False

        if v:
            is_tuple = True
        else:
            v = k
            is_tuple = False

        if v_is_path_or_intrinsic:
            if v.startswith("$"):  # It's a path
                v = apply_path(input, context, v)
            else:  # It's an Intrinsic Function
                v = evaluate_intrinsic_function(v)

        if is_tuple:
            return k, v
        else:
            return v

    def clone(template):
        """
        Recursively crawl the source JSON template creating a clone of its
        structure but evaluating and expanding fields whose name ends with “.$”
        """
        if isinstance(template, list):
            target = []
            for item in template:
                if isinstance(item, (dict, list)):
                    target.append(clone(item))
                else:
                    target.append(evaluate(item))
        elif isinstance(template, dict):
            target = {}
            for k, v in template.items():
                if isinstance(v, (dict, list)):
                    target[k] = clone(v)
                else:
                    k, v = evaluate(k, v)
                    target[k] = v
        return target

    if template == None or template == "":
        return input
    elif template == {}:
        return {}
    else:
        return clone(template)




#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""
This is a deliberately trivial logger implementation. As per 12FA it is treating
logs as an event stream, in this case stderr, see https://12factor.net/logs
A twelve-factor app never concerns itself with routing or storage of its output
stream. It should not attempt to write to or manage logfiles. Instead, each
running process writes its event stream, unbuffered, to stdout (or stderr).
"""

import json
import sys
assert sys.version_info >= (3, 0)  # Bomb out if not running Python3


import os, logging
import logging.config
from logging.handlers import RotatingFileHandler
import structlog

def inject_context(logger, method_name, event_dict):
    # inject current structlog context to stdlib logger calls from dependencies
    context_class = structlog.get_config().get("context_class")
    if context_class:
        context = context_class()
        # context object is not always subscriptable, so create list of kv pairs instead
        kv_pairs = [(k, context.get(k)) for k in context.keys()]
        event_dict.update(kv_pairs)
    return event_dict


# Use these processors for structlog and stdlib loggers
timestamper = structlog.processors.TimeStamper(fmt="iso",key="@timestamp")
shared_processors = [
    inject_context,
    structlog.stdlib.add_logger_name,
    structlog.stdlib.add_log_level,
    timestamper,
    structlog.stdlib.PositionalArgumentsFormatter(),
    structlog.processors.StackInfoRenderer(),
    structlog.processors.format_exc_info,
    structlog.processors.UnicodeDecoder(),
]


def configure_structlog():

    structlog.configure(
        processors=[structlog.stdlib.filter_by_level]
        + shared_processors
        + [structlog.stdlib.ProcessorFormatter.wrap_for_formatter],
        context_class=structlog.threadlocal.wrap_dict(dict),
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )

def get_structlog_formatter(*args, **kwargs):
    """ 
     The relevant docs for this aren't particularly clear, but it's possible 
     to specify a function that can return a Formatter object in a python 
     logging config file, instead of one of the built-in offerings. 
     
     In this case the arguments to that function default to () and {} 
     (ie. an empty tuple and dict), and so this function should have a 
     signature like def formatter_func(*args, **kwargs). We use the 
     get_structlog_formatter function like this to ensure that we can use
     structlog when configuring via a file and when not.
     
     Link to the relevant docs, I had to infer the above about the default arguments - 
     https://docs.python.org/3/library/logging.config.html#configuration-file-format
    """

    # Intentionally using stdlib JSON here as ujson.dumps doesn't accept all the 
    # named arguments that structlog expects. This is fine as the ASL Engine doesn't
    # log overly much anyway
    formatter = structlog.stdlib.ProcessorFormatter(
        processor=structlog.processors.JSONRenderer(json.dumps),
        foreign_pre_chain=shared_processors,
    )
    return formatter

def init_logging(log_name, log_level=logging.INFO):
    """
    Create a logger to use

    :param log_name: Name of log, usually processor name or similar
    :type log_name: str
    :return: Logger to use
    """
    logger = logging.getLogger(log_name)

    # If logger already has handlers just return it as it is already initialised
    if logger.hasHandlers():
        return logger

    # If the LOG_LEVEL environment variable is set use it to set the log level.
    log_levels = {
        "DEBUG": logging.DEBUG,
        "INFO": logging.INFO,
        "WARN": logging.WARN,
        "ERROR": logging.ERROR,
        "CRITICAL": logging.CRITICAL,
    }

    configured_level = os.environ.get("LOG_LEVEL", "INFO").upper()
    if configured_level in log_levels:
        log_level = log_levels.get(configured_level, logging.INFO)

    # Select automation friendly structured logging or more "human readable"
    # logging based on the value of the USE_STRUCTURED_LOGGING environment var.
    use_structured_logging = os.environ.get("USE_STRUCTURED_LOGGING", "false").lower() == "true"
    if use_structured_logging:
        configure_structlog()
    
    # Allows configuing the logger via an INI format configuration file
    # If a configuration file isn't supplied, we define sensible defaults in code
    # https://docs.python.org/3/library/logging.config.html#logging.config.fileConfig
    log_config_file = os.environ.get("LOG_CONFIG_FILE", "")
    if os.path.isfile(log_config_file):
        logging.config.fileConfig(log_config_file, disable_existing_loggers=False)

    else:
        if use_structured_logging:
            formatter = get_structlog_formatter()
        else:
            formatter = logging.Formatter(
                "[%(asctime)s] %(levelname)-8s - %(name)-15s : %(message)s"
            )

        handler = logging.StreamHandler()
        handler.setFormatter(formatter)

        logger.addHandler(handler)
        logger.setLevel(log_level)

    logger.debug("DEBUG enabled")
    return logger



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""
This creates data stores for ASL state machines (and their associated metadata)
and for execution metadata and execution history. The stores have dictionary
semantics where the ASL store is indexed by by the ARN of the ASL State Machine.
The format of the stored ASL dict objects is the same as DescribeStateMachine.
https://docs.aws.amazon.com/step-functions/latest/apireference/API_DescribeStateMachine.html.

The execution metadata is indexed by by the ARN of the execution.
The format of the stored execution dict objects is the same as DescribeExecution.
https://docs.aws.amazon.com/step-functions/latest/apireference/API_DescribeExecution.html

The execution history is also indexed by by the ARN of the execution.
The format of the stored execution history list objects is the same as that
of the "events" list from GetExecutionHistory and each event is an *immutable*
dict of the form illustrated in the GetExecutionHistory documentation.
https://docs.aws.amazon.com/step-functions/latest/apireference/API_GetExecutionHistory.html
"""

import sys
assert sys.version_info >= (3, 0)  # Bomb out if not running Python3

import json, os, threading, time, uuid, weakref
from collections import OrderedDict
from collections.abc import Mapping, MutableMapping, Sequence

from asl_workflow_engine.logger import init_logging


class JSONStore(MutableMapping):
    """
    https://stackoverflow.com/questions/3387691/how-to-perfectly-override-a-dict

    This class implements dict semantics, but uses a (very) simple JSON store.
    The approach is simple and in particular writes are inefficient as the
    approach to simply to serialise the store to JSON and write the JSON to
    the json_store file. This store is mainly intended to be used in basic
    scenarios and single instance test cases. For production scenarios it is
    recommended to use the Redis backed store, though that has more dependencies.
    """

    def __init__(self, json_store_name, *args, **kwargs):
        self.json_store = json_store_name

        self.logger = init_logging(log_name="asl_workflow_engine")
        self.logger.info("Creating JSONStore: {}".format(self.json_store))

        try:
            with open(self.json_store, "r") as fp:
                self.store = json.load(fp)
            self.logger.info(
                "JSONStore loading: {}".format(self.json_store)
            )
        except IOError as e:
            self.store = {}
        except ValueError as e:
            self.logger.warning(
                "JSONStore {} does not contain valid JSON".format(
                    self.json_store
                )
            )
            self.store = {}

        self.update(*args, **kwargs)  # use the free update to set keys

    def __str__(self):
        return str(self.store)

    def __getitem__(self, key):
        return self.store[key]

    def __setitem__(self, key, value):
        self.store[key] = value
        self._update_store()

    def __delitem__(self, key):
        del self.store[key]
        self._update_store()

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    #def update(self, iterable=tuple(), **kwargs):  # Uses "free" built-in update

    # From collections.abc.Mapping:
    def __contains__(self, key):
        return key in self.store

    def _update_store(self):
        # Serialise store to JSON and write the JSON to the json_store file.
        try:
            with open(self.json_store, "w") as fp:
                json.dump(self.store, fp)
            self.logger.info("Updating JSONStore: {}".format(self.json_store))
        except IOError as e:
            raise

    def set_ttl(self, key, ttl):
        pass

    def get_cached_view(self, key, default=None):
        return self.get(key, default)


class SimpleStore(dict):
    """
    Just a dict with trivial do nothing set_ttl and get_cached_view methods.
    """
    def set_ttl(self, key, ttl):
        pass

    def get_cached_view(self, key, default=None):
        return self.get(key, default)


class RedisStore(MutableMapping):
    """
    RedisStore is a Redis backed container implementing dict semantics.
    It is intended to be subclassed by RedisDictStore and RedisListStore as
    the use case is for storing a dict of dicts or a dict of lists.

    For a basic dict or list a Redis hash or list would suffice and the Pottery
    library gives those dict or list semantics. It is even possible to store
    a dict or list inside a Pottery RedisDict or RedisList. The issue is that
    those objects are (trasparently) JSON serialised. That is potentially not
    an issue for small and infrequently used items, but for objects that change
    a lot or could grow quite large, such as lists, the serialisation costs is
    O(2*N) *for every change* as each change to the object requires
    deserialisation->modification->serialisation.

    RedisStore basically just indexes the Redis hash or list objects (wrapped
    in Pottery RedisDict or RedisList. Redis itself actually indexes those,
    but this class applies an appropriate key prefix to disambiguate/namespace
    them and wraps the accesses in dict semantics so applications can access
    the RedisStore as if it were a dict of dicts or dict of lists, minimising
    any changes needed to the application in order to use the store.

    The reason we have RedisDictStore and RedisListStore is mainly because
    Redis does not allow empty hashes or lists, so if an application were to
    store say an empty list when it came to accessing that there would be no
    way of knowing its type without storing additional information. There are
    lots of ways to do that, but they all require additional writes and reads
    to the Redis server. As most applications are likely to want a dict of dicts
    or a dict of lists rather than a dict of dicts or lists it seems better to
    create specialisations rather than a more general form with additional cost.

    RedisStore makes use of Redis Server Assisted Client Side Caching where
    available (requires Redis > 6.0.0). This is enabled by accessing an object
    via the get_cached_view() method. The intended use case is again to mitigate
    deserialisation costs for objects accessed frequently. An example might be
    where we have JSON objects stored in a dict that we store in RedisStore.
    The get_cached_view() method will retrieve the object and deserialise it
    and store the *deserialised* object in an LRU cache. If that object is
    modified Redis will send a cache invalidation message and the object will be
    removed from the cache.
    """

    def get_connection(full_url, logger):
        """
        Class method to create or return the Redis connection. Defers imports
        of Redis libraries until we actually attempt a Redis connection.
        Use class method as we want all RedisStore instances to share connection.
        """
        try:
            return RedisStore.connection
        except AttributeError:
            """
            get_connection() supports URLs of the form:
            redis://localhost:6379?connection_attempts=20&retry_delay=10
            Redis URLs don't actually have a connection_attempts/retry_delay
            but we add them as they are convenient and consistent with AMQP URLs
            """
            split = full_url.split("?")  # Get query part
            url = split[0]  # The main URL before the ?
            # Use list comprehension to create options dict by splitting on & then =
            options = dict([] if len(split) == 1 else [
                i.split("=") for i in split[1].split("&")
            ])

            logger.info("RedisStore: Opening Connection to {}".format(url))

            # https://pypi.org/project/redis/
            # Use https://github.com/brainix/pottery for more Pythonic redis access
            from redis import Redis, ConnectionPool   # pip3 install redis
            from pottery import RedisDict, RedisList  # pip3 install pottery
            RedisStore.Redis = Redis
            RedisStore.RedisDict = RedisDict
            RedisStore.RedisList = RedisList

            # Defaults are the same defaults that Pika uses for AMQP connections.
            connection_attempts = int(options.get("connection_attempts", "1"))
            retry_delay = float(options.get("retry_delay", "2.0"))

            for i in range(connection_attempts):
                RedisStore.connection = Redis.from_url(url)
                try:
                    RedisStore.connection.ping()  # Check connection has succeeded
                    return RedisStore.connection
                except Exception as e:
                    err = e
                    logger.warning("RedisStore: {} retrying".format(e))
                    del RedisStore.connection
                time.sleep(retry_delay)

            logger.error("RedisStore: {} connection_attempts exceeded".format(err))
            sys.exit(1)

    def __init__(self, url, key="", cache_size=128, daemon=False):
        """
        key is used as a prefix/namespace in the Redis keyspace, so if we create
        multiple instances of a RedisStore then using different prefix keys for
        each instance will prevent odd result if we happen to need to store items
        with the same index keys in different stores.

        cache_size is the maximum number of items to store in the LRU cache used
        by get_cached_view(). Setting cache_size to 0 will disable the cache, so
        calls to get_cached_view() will behave exactly like get() or [].

        daemon is used to set the thread listening for cache invalidation
        messages from the Redis server as a daemon or not.
        """
        self.logger = init_logging(log_name="asl_workflow_engine")
        self.logger.info(
            "Creating {} {}: {}".format(key, self.__class__.__name__, url)
        )

        self.redis = None  # In case get_connection() fails, which will __del__ on exit.
        self.redis = RedisStore.get_connection(url, self.logger)

        # Get Redis version as an int, so 6.0.0 would be 600 to aid version tests.
        redis_version = self.redis.info("server").get("redis_version", "0")
        self.redis_version = int(redis_version.replace(".", ""))
        if self.redis_version < 600:
            self.logger.warning("Server assisted client side caching is not "
                                "supported in Redis version {}, read performance "
                                "might be reduced.".format(redis_version))

        self.key = key

        # Enabling the cache is deferred until first call to get_cached_view().
        self.cache = None
        self.cache_size = cache_size
        self.daemon = daemon
        self.tracker_id = None

    def __del__(self):
        self.stop()

    def _start_tracking(self):
        """
        Enable Redis server-assisted client side caching. Requires Redis >= 6.0.0
        https://redis.io/topics/client-side-caching
        https://engineering.redislabs.com/posts/redis-assisted-client-side-caching-in-python/
        """

        """
        Redis serialisation protocol 2 (RESP2) currently used in redis-py lacks
        the RESP3 "push" capability, so Redis server-assisted client side caching
        also broadcasts invalidation messages using the existing PubSub support.
        RESP2 lacks the ability to multiplex different kind of information in
        the same connection so we create a separate tracker subscriber connection
        from the same connection pool as the main Redis connection.
        """
        self.tracker = RedisStore.Redis(connection_pool=self.redis.connection_pool)

        """
        It is important to retrieve a copy of self.tracker.client_id() here
        as when we later call pubsub.subscribe() that results in a new logical
        connection, so if we were to call tracker.client_id() again it would
        have a different value. This is a subtle gotcha, as it is tempting to
        use self.tracker.client_id() in the CLIENT TRACKING ON call which
        would cause a redirect *to the wrong logical client*.
        """
        self.tracker_id = self.tracker.client_id()

        """
        In the following code the obvious way to register the handler is using:
        pubsub.subscribe(**{"__redis__:invalidate": self._cache_invalidation_handler})
        That works, but results in a circular reference with the tracker_thread.
        The circular reference means the RedisStore won't get garbage collected,
        so instead we create a weakref to self then use a lambda as the
        invalidate handler which, when called, will dereference the weakref
        and call the real _cache_invalidation_handler. Now when del is called
        on a RedisStore instance everything should be GC'd and released.
        https://stackoverflow.com/questions/2436302/when-to-use-weak-references-in-python

        Another gotcha is the pubsub.run_in_thread() call. It's easy to assume
        that, like many similar subscription APIs, the subscription thread will
        *block* (and thus consume no CPU) until a message is sent from the server.
        Unfortunately that's not the case and, as may be seen from the source,
        https://github.com/andymccurdy/redis-py/blob/master/redis/client.py#L3708
        the thread simply calls get_message in a loop. Setting a sleep_time
        trades latency for CPU, and with 10ms latency the CPU used is negligible
        but it's still a bit ugly to have a busy-wait.
        """
        weakref_to_self = weakref.ref(self)
        pubsub = self.tracker.pubsub(ignore_subscribe_messages=True)
        pubsub.subscribe(**{
            "__redis__:invalidate": lambda message:
                weakref_to_self()._cache_invalidation_handler(message)
        })
        #self.tracker_thread = pubsub.run_in_thread(sleep_time=0.01, daemon=self.daemon)

        """
        Rather than use pubsub.run_in_thread() as above, the following block runs
        pubsub.listen() in a thread. It's a slightly more complex approach, but
        in this case the thread blocks rather than busy-waiting, though that
        adds additional complexity in the stop() method too. Again we use a
        weakref to the pubsub object rather than the more obvious self.pubsub
        to avoid a circular reference that would prevent garbage collection.
        """
        self.weakref_to_pubsub = weakref.ref(pubsub)
            
        def pubsub_listener():
            for message in pubsub.listen():
                """
                pubsub.listen() will block until there is a message available.
                If there is a message handler it will call that rather than
                yield. In general this means that this call will *appear* to do
                nothing and won't yield. To cleanly exit it is necessary to
                remove the message handler and do a basic subscribe, then
                publish a dummy message that will cause pubsub.listen() to
                yield. See the stop() method for how to cleanly unblock and exit.
                """
                break

        self.tracker_thread = threading.Thread(
            target=pubsub_listener,
            daemon=self.daemon,
        )
        self.tracker_thread.start()
        
        """
        Switch on tracking. We need to use execute_command() here as redis-py
        doesn't currently support it via a higher-level API call.
        """
        self.redis.execute_command(
            "CLIENT", "TRACKING", "ON", "REDIRECT", self.tracker_id
        )

    def _cache_invalidation_handler(self, message):
        """
        Handles key invalidation messages sent by the Redis server.
        """
        keys = message["data"]  # This will contain an array of invalidated keys.        
        for k in keys:
            # Keys are passed as an array of binary strings, with prefixes.
            key = self._remove_prefix(k.decode("utf-8"))
            if key in self.cache:
                del self.cache[key]

    def _remove_prefix(self, key):
        """
        Boiler plate helper method to remove the internal prefixes used in the
        Redis keys and return the keys as used by applications using RedisStore.
        """
        prefix = self.key + ":"
        if key.startswith(prefix):
            return key[len(prefix):]
        return key

    def __delitem__(self, key):
        k = self.key + ":" + key
        self.redis.delete(k)

    def __len__(self):
        """
        Warning. Because this class is storing Redis data structures (hashes and
        lists) the keys used in RedisStore are prefixed Redis database keys.
        There is no Redis API call to get a count of keys with a given prefix,
        so we use scan with match set to the prefix. This works, but be aware
        that a call to len will have O(N) time complexity and not O(1).
        This shouldn't be an issue as the use cases for RedisStore in this
        application shouldn't need to call len, but it's important to be aware.
        """
        count = 0
        cursor = "0"  # Not 0 or the while loop will immediately terminate.
        while cursor != 0:
            cursor, iterable = self.redis.scan(cursor=cursor, match=self.key + ":*")
            count = count + len(iterable)
        return count

    def __iter__(self):
        """
        Generator function that yields results from a Redis scan. The yielded
        results are converted from the binary strings returned by scan and have
        their prefixes stripped, so this call behaves semantically like iterating
        a dict yielding the dict keys.
        """
        cursor = "0"  # Not 0 or the while loop will immediately terminate.
        while cursor != 0:
            cursor, iterable = self.redis.scan(cursor=cursor, match=self.key + ":*")
            yield from (self._remove_prefix(value.decode("utf-8")) for value in iterable)

    # From collections.abc.Mapping:
    def __contains__(self, key):
        k = self.key + ":" + key
        return bool(self.redis.exists(k))

    def set_ttl(self, key, ttl):
        k = self.key + ":" + key
        self.redis.expire(k, ttl)

    def _write_to_cache(self, key, value):
        """
        The cache is *eventually consistent* and currently the only place
        _write_to_cache() is called is when a cache read fails in
        get_cached_view(). The implication of this is that even for *local*
        writes to the store the new data won't be available in the cache until
        after the server has sent the cache invalidation message.
        It would be trivial to add a _write_to_cache() call in the __setitem__
        calls of RedisDictStore and RedisListStore, which would make local
        updates available immediately in the cache for the case where
        applications might write a whole dict or list to the store, but in
        general clients might do something like the following:
        asl_store[state_machine_arn]["type"] = "EXPRESS"
        Under the hood that *actually* writes to the pottery RedisDict object.
        It would be possible to interpose a call to _write_to_cache() onto
        the pottery RedisDict and RedisList mutator methods, but on balance
        eventually consistent semantics should be good enough though it is 
        important to be aware of the subtlety.
        """
        if self.cache is None:  # Only write to cache if it is enabled.
            return value

        # Store *deserialised* dict or list objects in cache.
        if isinstance(value, RedisStore.RedisDict):
            cached_value = dict(value)
        elif isinstance(value, RedisStore.RedisList):
            cached_value = list(value)
        else:
            cached_value = value

        self.cache[key] = cached_value  # Write to cache
        # Implement LRU semantics e.g. evict the oldest item from cache
        if len(self.cache) > self.cache_size:
            oldest = next(iter(self.cache))
            del self.cache[oldest]
        return cached_value

    def get_cached_view(self, key, default=None):
        """
        We only need to enable server-assisted client side caching if we are
        actually using caching. In general caching is mainly useful here if we
        are doing many reads of infrequently updated immutable objects, so we
        can amortise the deserialisation costs in addition to minimising the
        network round-trips of Redis GET calls. If objects are read infrequently
        or the dicts or lists being accessed need to be modified then use the
        basic get() or [] operator. This method returns a *native* dict or list
        and not a RedisDict or RedisList, so any changes wil not be reflected
        in the underlying Redis store.
        """

        """
        Older Redis versions don't support RSACSC so just return uncached object
        for now. TODO we could have producers publish when they make changes,
        e.g. self.redis.publish("__redis__:invalidate", "test1")
        That should work, though will be less efficient as producers and
        consumers are decoupled, so every change might need published and the
        increased costs might well negate the benefits of caching.
        """
        if self.redis_version < 600 or self.cache_size == 0:
            return self[key]

        if self.tracker_id is None:
            self.cache = OrderedDict()
            self._start_tracking()

        try:  # Get from local cache
            value = self.cache[key]
            self.cache.move_to_end(key)
        except KeyError:  # Get from Redis server
            value = self[key]
            value = self._write_to_cache(key, value)

        return value

    def stop(self):
        """
        Tidy up resources. This will also be called by the destructor, so it
        shouldn't be necessary to call this explicitly.
        """
        if self.redis is not None and self.tracker_id is not None:
            self.redis.execute_command("CLIENT", "TRACKING", "OFF")

            try: # If subscribing is implemented using pubsub.run_in_thread()
                self.tracker_thread.stop()
            except AttributeError:
                """
                If subscribing is implemented using pubsub.listen() in a worker
                thread we remove the message handler by calling subscribe()
                again, but with no handler for the channel. We then publish to
                the channel, which will now cause pubsub.listen() to yield as
                there is no listener. Allowing the yield to succeed causes the
                worker thread to exit.
                """
                pubsub = self.weakref_to_pubsub()
                pubsub.subscribe("__redis__:invalidate")
                self.redis.publish("__redis__:invalidate", "exit")
                self.tracker_thread.join()

            self.tracker.close()
            self.tracker_id = None
            self.redis.close()


class RedisDictStore(RedisStore):
    """
    https://stackoverflow.com/questions/3387691/how-to-perfectly-override-a-dict

    This class implements dict of dicts semantics. Note that whilst the dicts
    stored in a RedisDictStore are themselves mutable, stored as Redis hashes,
    any objects that they contain are serialised and immutable, though they
    may be overwritten by a new serialised object.
    """
    def __getitem__(self, key):
        k = self.key + ":" + key
        return RedisStore.RedisDict(redis=self.redis, key=k)

    def __setitem__(self, key, value):
        k = self.key + ":" + key

        if isinstance(value, RedisStore.RedisDict) and value.key == k:
            pass  # If setting the same RedisDict instance we needn't do anything more.
        elif isinstance(value, Mapping):
            # Delete then store seems the only way to replace entire Redis hash.
            self.redis.delete(k)
            if value:
                RedisStore.RedisDict(value, redis=self.redis, key=k)
        else:
            raise TypeError("RedisDictStore can only store items of type Mapping")


class RedisListStore(RedisStore):
    """
    https://stackoverflow.com/questions/3387691/how-to-perfectly-override-a-dict

    This class implements dict of lists semantics. Note that whilst the lists
    stored in a RedisListStore are themselves mutable, stored as Redis lists,
    any objects that they contain are serialised and immutable, though they
    may be overwritten by a new serialised object.
    """
    def __getitem__(self, key):
        k = self.key + ":" + key
        return RedisStore.RedisList(redis=self.redis, key=k)

    def __setitem__(self, key, value):
        k = self.key + ":" + key
        
        if isinstance(value, RedisStore.RedisList) and value.key == k:
            pass  # If setting the same RedisList instance we needn't do anything more.
        elif isinstance(value, Sequence):
            # Delete then store seems the only way to replace entire Redis list.
            self.redis.delete(k)
            if value:
                RedisStore.RedisList(value, redis=self.redis, key=k)
        else:
            raise TypeError("RedisListStore can only store items of type Sequence")


def create_ASL_store(store_url):
    """
    The ASL store is semantically a dict of dicts indexed by by the ARN of the
    ASL State Machine and returning dicts in the format of DescribeStateMachine.
    """
    if store_url.startswith("redis://"):
        # TODO should the LRU cache size be configurable? Probably OK for now.
        return RedisDictStore(store_url, "asl_store", cache_size=1024, daemon=True)
    else:
        return JSONStore(store_url)


def create_executions_store(store_url):
    """
    The executions store is semantically a dict of dicts indexed by by the ARN
    of the execution and returning dicts in the format of DescribeExecution.
    The DISABLE_EXECUTIONS_STORE environment variable is mainly intended to be
    used during performance testing to gauge the performance impact of RedisStore.
    """
    if (store_url.startswith("redis://") and
        os.environ.get("DISABLE_EXECUTIONS_STORE", "false").lower() != "true"):
        return RedisDictStore(store_url, "executions")
    else:
        return SimpleStore()


def create_history_store(store_url):
    """
    The history store is semantically a dict of lists indexed by by the ARN of
    the execution and returning lists in the format of the "events" list from 
    GetExecutionHistory. Each event stored in the lists is an *immutable*
    dict of the form illustrated in the GetExecutionHistory documentation.
    The DISABLE_HISTORY_STORE environment variable is mainly intended to be
    used during performance testing to gauge the performance impact of RedisStore.
    """
    if (store_url.startswith("redis://") and
        os.environ.get("DISABLE_HISTORY_STORE", "false").lower() != "true"):
        return RedisListStore(store_url, "execution_history")
    else:
        return SimpleStore()




#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
"""
Defines the exceptions relating to ASL itself as defined in
https://states-language.net/spec.html.
"""

import sys
assert sys.version_info >= (3, 0)  # Bomb out if not running Python3


class Timeout(Exception):
    pass


class TaskFailed(Exception):
    pass


class Permissions(Exception):
    pass


class ResultPathMatchFailure(Exception):
    pass


class ParameterPathFailure(Exception):
    pass


class IntrinsicFailure(Exception):
    pass


class BranchFailed(Exception):
    pass


class NoChoiceMatched(Exception):
    pass


# Not defined in the ASL spec but used in the Choice state path handling.
class PathMatchFailure(Exception):
    pass
